/* aloreapi.c - Alore C API

   Copyright (c) 2010-2011 Jukka Lehtosalo

   Alore is licensed under the terms of the MIT license.  See the file
   LICENSE.txt in the distribution.
*/

/* Alore C API functions: These functions along with other functions and macros
   defined in alore.h form the public Alore extension API. Functions, macros,
   types and constants not defined in alore.h are internal and should not be
   used by extensions if compatibility with future Alore versions is
   desired. */

#include <math.h>
#include <stdlib.h>
#include "alore.h"
#include "runtime.h"
#include "class.h"
#include "memberid.h"
#include "int.h"
#include "float.h"
#include "str.h"
#include "array.h"
#include "symtable.h"
#include "parse.h"
#include "std_module.h"
#include "gc.h"
#include "mem.h"
#include "internal.h"
#include "thread_athread.h"


static AValue InstanceOp(AThread *t, AValue a, AValue b, int member);


void ASetConstGlobalByNum(AThread *t, int num, AValue v)
{
    if (!ASetConstGlobalValue(t, num, v))
        ARaiseMemoryError(t);
}


int AGetInt(AThread *t, AValue val)
{
#if A_VALUE_BITS == 32 && A_LONG_INT_DIGIT_BITS == 16
    /* IDEA: Use AGetLongIntDigit and other API macros instead of LongInt *. */
    if (AIsShortInt(val))
        return AValueToInt(val);
    else if (AIsLongInt(val)) {
        ALongInt *li = AValueToLongInt(val);
        ALongIntDoubleDigit dd = li->digit[0] + (li->digit[1] << 16);
        if ((AGetLongIntSign(li) && dd > (1U << 31)) ||
            (!AGetLongIntSign(li) && dd > INT_MAX) ||
            AGetLongIntLen(li) > 2)
            return ARaiseValueError(t, AMsgOutOfRange);
        return AGetLongIntSign(li) ? -dd : dd;
    } else
        return ARaiseTypeError(t, AMsgIntExpected);
#elif A_VALUE_BITS == 64
    if (AIsShortInt(val)) {
        ASignedValue v = AValueToInt(val);
        if (v > 0x7fffffff || v < -0x80000000L)
            ARaiseValueError(t, AMsgOutOfRange);
        return (int)v;
    } else if (AIsLongInt(val))
        return ARaiseValueError(t, AMsgOutOfRange);
    else
        return ARaiseTypeError(t, AMsgIntExpected);
#else
    /* Not supported */
    error!;
#endif
}


unsigned AGetIntU(AThread *t, AValue val)
{
#if A_VALUE_BITS == 32 && A_LONG_INT_DIGIT_BITS == 16
    if (AIsShortInt(val)) {
        if ((ASignedValue)val < 0)
            return ARaiseValueError(t, AMsgNegativeValueNotExpected);
        return AValueToInt(val);
    } else if (AIsLongInt(val)) {
        if (AIsNegLongInt(val) || ALongIntLen(val) > 2)
            return ARaiseValueError(t, AMsgOutOfRange);
        return AGetLongIntDigit(val, 0) + (AGetLongIntDigit(val, 1) << 16);
    } else
        return ARaiseTypeError(t, AMsgIntExpected);
#elif A_VALUE_BITS == 64
    if (AIsShortInt(val)) {
        ASignedValue v = AValueToInt(val);
        if (v < 0 || v > 0xffffffff)
            return ARaiseValueError(t, AMsgOutOfRange);
        return (int)v;
    } else if (AIsLongInt(val))
        return ARaiseValueError(t, AMsgOutOfRange);
    else
        return ARaiseTypeError(t, AMsgIntExpected);
#else
    /* Not supported */
    error!;
#endif
}


AInt64 AGetInt64(AThread *t, AValue val)
{
#if A_VALUE_BITS == 32 && A_LONG_INT_DIGIT_BITS == 16
    if (AIsShortInt(val))
        return AValueToInt(val);
    else if (AIsLongInt(val)) {
        AInt64 i;
        int n;
        
        if (ALongIntLen(val) > 4
            || (ALongIntLen(val) == 4 &&
                (AGetLongIntDigit(val, 3) > 0x8000)))
            return ARaiseValueError(t, AMsgOutOfRange);
        if (ALongIntLen(val) == 4 && AGetLongIntDigit(val, 3) == 0x8000) {
            n = 2;
            while (n > 0 && AGetLongIntDigit(val, n) == 0)
                n--;
            if (AGetLongIntDigit(val, n) > 0 || !AIsNegLongInt(val))
                return ARaiseValueError(t, AMsgOutOfRange);
        }
        
        i = 0;
        for (n = ALongIntLen(val) - 1; n >= 0; n--)
            i = (i << 16) + AGetLongIntDigit(val, n);
        return AIsNegLongInt(val) ? -i : i;
    } else
        return ARaiseTypeError(t, AMsgIntExpected);
#elif A_VALUE_BITS == 64 && A_LONG_INT_DIGIT_BITS == 32
    if (AIsShortInt(val))
        return AValueToInt(val);
    else if (AIsLongInt(val)) {
        AInt64 i;
        int n;
        
        if (ALongIntLen(val) > 2
            || (ALongIntLen(val) == 2 &&
                (AGetLongIntDigit(val, 1) > 0x80000000U)))
            return ARaiseValueError(t, AMsgOutOfRange);
        if (ALongIntLen(val) == 2 && AGetLongIntDigit(val, 1) == 0x80000000U) {
            if (AGetLongIntDigit(val, 0) > 0 || !AIsNegLongInt(val))
                return ARaiseValueError(t, AMsgOutOfRange);
        }
        
        i = 0;
        for (n = ALongIntLen(val) - 1; n >= 0; n--)
            i = (i << 32) + AGetLongIntDigit(val, n);
        return AIsNegLongInt(val) ? -i : i;
    } else
        return ARaiseTypeError(t, AMsgIntExpected);
#else
    /* Not supported */
    error!;
#endif
}


AIntU64 AGetIntU64(AThread *t, AValue val)
{
#if A_VALUE_BITS == 32 && A_LONG_INT_DIGIT_BITS == 16
    if (AIsShortInt(val)) {
        if ((ASignedValue)val < 0)
            return ARaiseValueError(t, AMsgNegativeValueNotExpected);
        return AValueToInt(val);
    } else if (AIsLongInt(val)) {
        AIntU64 i;
        int n;
        
        if (AIsNegLongInt(val) || ALongIntLen(val) > 4)
            return ARaiseValueError(t, AMsgOutOfRange);
        
        i = 0;
        for (n = ALongIntLen(val) - 1; n >= 0; n--)
            i = (i << 16) + AGetLongIntDigit(val, n);
        return i;
    } else
        return ARaiseTypeError(t, AMsgIntExpected);
#elif A_VALUE_BITS == 64 && A_LONG_INT_DIGIT_BITS == 32
    if (AIsShortInt(val)) {
        if ((ASignedValue)val < 0)
            return ARaiseValueError(t, AMsgNegativeValueNotExpected);
        return AValueToInt(val);
    } else if (AIsLongInt(val)) {
        AIntU64 i;
        int n;
        
        if (AIsNegLongInt(val) || ALongIntLen(val) > 2)
            return ARaiseValueError(t, AMsgOutOfRange);
        
        i = 0;
        for (n = ALongIntLen(val) - 1; n >= 0; n--)
            i = (i << 32) + AGetLongIntDigit(val, n);
        return i;
    } else
        return ARaiseTypeError(t, AMsgIntExpected);
#else
    /* Not supported */
    error!;
#endif
}


AValue AMakeInt(AThread *t, int i)
{
#if A_VALUE_BITS != 32 && A_VALUE_BITS != 64
    error! /* FIX */;
#endif
    
#if A_VALUE_BITS == 32
    /* 32-bit values */
    if (i <= A_SHORT_INT_MAX && i >= A_SHORT_INT_MIN)
        return AIntToValue(i);
    else {
        AValue v = ACreateLongIntFromIntND(t, i);
        if (AIsError(v))
            ADispatchException(t);
        return v;
    }
#else
    /* 64-bit values */
    return AIntToValue(i);
#endif
}


AValue AMakeIntU(AThread *t, unsigned i)
{
#if A_VALUE_BITS == 32 && A_LONG_INT_DIGIT_BITS == 16
    /* 32-bit values */
    if (i <= A_SHORT_INT_MAX)
        return AIntToValue(i);
    else {
        AValue v = AMakeLongInt(t, 2, FALSE);
        ASetLongIntDigit(v, 0, i & 0xffff);
        ASetLongIntDigit(v, 1, i >> 16);
        return v;
    }
#elif A_VALUE_BITS == 64
    /* 64-bit values */
    return AIntToValue(i);
#else
    /* Not supported */
    error!;
#endif
}


AValue AMakeInt64(AThread *t, AInt64 i)
{
#if A_VALUE_BITS == 32 && A_LONG_INT_DIGIT_BITS == 16
    if (i <= A_SHORT_INT_MAX && i >= A_SHORT_INT_MIN)
        return AIntToValue(i);
    else {
        AValue v = AMakeLongInt(t, 4, FALSE);
        ABool neg = i < 0;
        AInt64 ui = i < 0 ? -i : i;
        
        ASetLongIntDigit(v, 0, ui & 0xffff);
        ASetLongIntDigit(v, 1, (ui >> 16) & 0xffff);
        ASetLongIntDigit(v, 2, (ui >> 32) & 0xffff);
        ASetLongIntDigit(v, 3, (ui >> 48) & 0xffff);
        return ANormalize(t, AValueToLongInt(v), neg);
    }
#elif A_VALUE_BITS == 64 && A_LONG_INT_DIGIT_BITS == 32
    if (i <= A_SHORT_INT_MAX && i >= A_SHORT_INT_MIN)
        return AIntToValue(i);
    else {
        AValue v = AMakeLongInt(t, 2, FALSE);
        ABool neg = i < 0;
        AInt64 ui = i < 0 ? -i : i;
        
        ASetLongIntDigit(v, 0, ui & 0xffffffffU);
        ASetLongIntDigit(v, 1, ui >> 32);
        return ANormalize(t, AValueToLongInt(v), neg);
    }
#else
    /* Not supported */
    error!;
#endif
}


AValue AMakeIntU64(AThread *t, AIntU64 i)
{
#if A_VALUE_BITS == 32 && A_LONG_INT_DIGIT_BITS == 16
    if (i <= A_SHORT_INT_MAX)
        return AIntToValue(i);
    else {
        AValue v = AMakeLongInt(t, 4, FALSE);
        ASetLongIntDigit(v, 0, i & 0xffff);
        ASetLongIntDigit(v, 1, (i >> 16) & 0xffff);
        ASetLongIntDigit(v, 2, (i >> 32) & 0xffff);
        ASetLongIntDigit(v, 3, (i >> 48) & 0xffff);
        return ANormalize(t, AValueToLongInt(v), 0);
    }
#elif A_VALUE_BITS == 64
    if (i <= A_SHORT_INT_MAX)
        return AIntToValue(i);
    else {
        AValue v = AMakeLongInt(t, 2, FALSE);
        ASetLongIntDigit(v, 0, i & 0xffffffffU);
        ASetLongIntDigit(v, 1, i >> 32);
        return ANormalize(t, AValueToLongInt(v), 0);
    }
#else
    /* Not supported */
    error!;
#endif
}


AValue AMakeLongInt(AThread *t, int len, ABool isNeg)
{
    ALongInt *li = AAlloc(t, AGetLongIntSize(len));
    if (li == NULL)
        return ARaiseMemoryError(t);
    AInitLongIntBlock(li, len, isNeg);
    return ALongIntToValue(li);
}


double AGetFloat_(AThread *t, AValue v)
{
    if (AIsShortInt(v))
        return AValueToInt(v);
    else if (AIsLongInt(v))
        return ALongIntToFloat(v);
    else
        return ARaiseTypeError(t, AMsgFloatExpected);
}


AValue AMakeFloat(AThread *t, double f)
{
    /* IDEA: Optimize this. */
    AValue v = ACreateFloat(t, f);
    if (AIsError(v))
        ADispatchException(t);
    return v;
}


/* Determine whether f is IEEE 754 positive or negative infinity. */
ABool AIsInf(double f)
{
    return (f > 0.0 || f < 0.0) && f + f == f;
}


/* Determine whether f is IEEE NaN (Not-a-Number). */
ABool AIsNaN(double f)
{
    return f != f;
}


Assize_t AStrLenUtf8(AValue v)
{
    Assize_t i;
    Assize_t len;
    Assize_t utf8len;

    len = AStrLen(v);
    utf8len = 0;
    for (i = 0; i < len; i++)
        utf8len += A_UTF8_LEN(AStrItem(v, i));
    
    return utf8len;
}


Assize_t AGetStr(AThread *t, AValue s, char *buf, Assize_t bufLen)
{
    Assize_t len = AGetCStrFromString(t, s, buf, bufLen);
    if (len < 0)
        ADispatchException(t);
    return len;
}


Assize_t AGetStrW(AThread *t, AValue s, AWideChar *buf, Assize_t bufLen)
{
    Assize_t i;
    Assize_t len;
    
    if (!AIsStr(s))
        return ARaiseTypeError(t, AMsgStrExpected);

    len = AStrLen(s);
    if (len >= bufLen)
        return ARaiseValueError(t, "Str too long");
    for (i = 0; i < len; i++)
        buf[i] = AStrItem(s, i);
    buf[i] = '\0';
    return len;
}


Assize_t AGetStrUtf8(AThread *t, AValue s, char *buf, Assize_t bufLen)
{
    Assize_t i, j;
    Assize_t len;
    Assize_t resultLen;
    
    AExpectStr(t, s);

    len = AStrLen(s);

    /* Calculate the length of the UTF-8 encoded string. */
    resultLen = 0;
    for (i = 0; i < len; i++)
        resultLen += A_UTF8_LEN(AStrItem(s, i));
    
    if (resultLen >= bufLen)
        ARaiseValueError(t, "Str too long");

    /* UTF-8 encode the input string into buf. */
    j = 0;
    for (i = 0; i < len; i++) {
        AWideChar ch = AStrItem(s, i);
        if (ch <= 0x7f) {
            buf[j] = ch;
            j++;
        } else if (ch <= 0x7ff) {
            buf[j] = 0xc0 | (ch >> 6);
            buf[j + 1] = 0x80 | (ch & 0x3f);
            j += 2;
        } else {
            buf[j] = 0xe0 | (ch >> 12);
            buf[j + 1] = 0x80 | ((ch >> 6) & 0x3f);
            buf[j + 2] = 0x80 | (ch & 0x3f);
            j += 3;
        }
    }
    buf[j] = '\0';
    
    return resultLen;
}


AValue AMakeStr(AThread *t, const char *str)
{
    AValue v = ACreateString(t, str, strlen(str));
    if (AIsError(v))
        ADispatchException(t);
    return v;
}


AValue AMakeStrW(AThread *t, const AWideChar *str)
{
    AValue v;
    int len;

    for (len = 0; str[len] != '\0'; len++);
    v = ACreateWideString(t, str, len);
    if (AIsError(v))
        ADispatchException(t);
    return v;
}


AValue AMakeStrUtf8(AThread *t, const char *str)
{
    int i, j;
    int len;
    int resultLen;
    AValue s;

    len = strlen(str);    
    resultLen = 0;

    for (i = 0; i < len; ) {
        i += A_UTF8_SKIP((unsigned char)str[i]);
        resultLen++;
    }

    if (i != len)
        goto DecodeError;
    
    s = AMakeEmptyStrW(t, resultLen);
    j = 0;
    for (i = 0; i < resultLen; i++) {
        unsigned char ch1 = str[j];
        AWideChar result;
        
        if (ch1 <= 0x7f) {
            result = ch1;
            j++;
        } else if (ch1 < 0xc2) {
            j++;
            goto DecodeError;
        } else if (ch1 < 0xe0) {
            unsigned char ch2 = str[j + 1];
            j += 2;
            if (ch2 < 0x80 || ch2 > 0xbf)
                goto DecodeError;
            result = ((ch1 & 0x1f) << 6) | (ch2 & 0x3f);
        } else {
            unsigned char ch2 = str[j + 1];
            unsigned char ch3 = str[j + 2];
            j += 3;
            if (ch2 < 0x80 || ch2 > 0xbf || ch3 < 0x80 || ch3 > 0xbf
                || (ch1 == 0xe0 && ch2 < 0xa0))
                goto DecodeError;
            result = ((ch1 & 0xf) << 12) | ((ch2 & 0x3f) << 6) | (ch3 & 0x3f);
        }
        
        ASetStrItemW(s, i, result);
    }
    
    return s;

  DecodeError:

    return ARaiseValueError(t, "Invalid UTF-8 sequence");
}

    
AValue AMakeEmptyStr(AThread *t, Assize_t len)
{
    AValue v = ACreateString(t, NULL, len);
    if (AIsError(v))
        ADispatchException(t);
    return v;
}


AValue AMakeEmptyStrW(AThread *t, Assize_t len)
{
    AWideString *s = AAlloc(t, sizeof(AValue) + sizeof(AWideChar) * len);
    if (s == NULL) {
        ADispatchException(t);
        return AError;
    }
    AInitNonPointerBlock(&s->header, sizeof(AWideChar) * len);
    return AWideStrToValue(s);
}


AValue AToStr(AThread *t, AValue o)
{
    AValue *temp = AAllocTemp(t, o);
    AValue str = AStdStr(t, temp);
    AFreeTemp(t);
    return str;
}


AValue AConcat(AThread *t, AValue a, AValue b)
{
    AValue v;
    
    if (AIsStr(a)) {
        if (AIsStr(b))
            v = AConcatStrings(t, a, b);
        else
            return ARaiseTypeError(t, NULL);
    } else
        v = InstanceOp(t, a, b, AM_ADD);

    if (AIsError(v))
        ADispatchException(t);

    return v;
}


AValue ASubStr(AThread *t, AValue s, Assize_t i1, Assize_t i2)
{
    AValue v = ACreateSubStr(t, s, i1, i2);
    if (AIsError(v))
        ADispatchException(t);
    return v;
}


AWideChar AGetCh(AThread *t, AValue ch)
{
    if (AIsStr(ch) && AStrLen(ch) == 1)
        return AStrItem(ch, 0);
    else {
        ARaiseValueError(t, AMsgChrExpected);
        return 0; /* Not reached */
    }
}


AValue AMakeCh(AThread *t, AWideChar ch)
{
    /* IDEA: Return a narrow string if ch < 256. Consider optimizing for
             speed. */
    AWideChar s[2];
    s[0] = ch;
    s[1] = '\0';
    return AMakeStrW(t, s);
}


void ASetArrayItem(AThread *t, AValue a, Assize_t index, AValue v)
{
    ABool result;
    AInstance *i;
    AFixArray *aa;

    *t->tempStack = v;
    i = AValueToInstance(a);
    aa = AValueToFixArray(i->member[A_ARRAY_A]);
    AModifyObject_M(t, &aa->header, aa->elem + index, t->tempStack, result);
    *t->tempStack = AZero;
    if (!result)
        ARaiseMemoryError(t);
}


AValue AMakeArray(AThread *t, Assize_t len)
{
    AValue v = AMakeArrayND(t, len);
    if (AIsError(v))
        ADispatchException(t);
    return v;
}


AValue ASubArray(AThread *t, AValue a, Assize_t i1, Assize_t i2)
{
    AValue sub;
    AExpectArray(t, a);
    sub = ASubArrayND(t, a, i1, i2);
    if (AIsError(sub))
        ADispatchException(t);
    return sub;
}


void ASetFixArrayItem(AThread *t, AValue a, Assize_t index, AValue v)
{
    ABool result;
    AFixArray *aa;

    *t->tempStack = v;
    aa = AValueToFixArray(a);
    AModifyObject_M(t, &aa->header, aa->elem + index, t->tempStack, result);
    *t->tempStack = AZero;
    if (result)
        return;
    else
        ARaiseMemoryError(t);
}


AValue *AAllocTemps(AThread *t, int n)
{
    int i;
    if (!AAllocTempStack_M(t, n))
        ADispatchException(t);
    for (i = 1; i <= n; i++)
        *(t->tempStackPtr - i) = AZero;
    return t->tempStackPtr - n;
}


ABool AIsValue(AThread *t, AValue v, AValue type)
{
    AIsResult res = AIsOfType(v, type);
    if (res == A_IS_TRUE)
        return 1;
    else if (res == A_IS_FALSE)
        return 0;
    else {
        ARaiseTypeError(t, AMsgTypeExpected);
        return 0;
    }
}


ABool AIsType(AValue v)
{
    /* The thread parameter may by NULL since we know that the last argument
       is a type. */
    return AIsValue(NULL, v, AGlobalByNum(AStdTypeNum));
}


AValue ARaise(AThread *t, const char *type, const char *msg)
{
    int num = AGetGlobalNum(t, type);
    ARaiseByNum(t, num, msg);
    ADispatchException(t);
    return AError;
}


AValue ARaiseByType(AThread *t, AValue type, const char *msg)
{
    if (!AIsNonSpecialType(type))
        ARaiseTypeError(t, AMsgTypeExpected);
    ARaiseByNum(t, AValueToType(type)->sym->num, msg);
    ADispatchException(t);
    return AError;
}


AValue ARaiseValue(AThread *t, AValue v)
{
    ARaiseExceptionValue(t, v);
    ADispatchException(t);
    return AError;
}


AValue ARaiseMemoryError(AThread *t)
{
    ARaiseMemoryErrorND(t);
    ADispatchException(t);
    return AError;
}


AValue ARaiseTypeError(AThread *t, const char *msg)
{
    ARaiseByNum(t, AErrorClassNum[EX_TYPE_ERROR], msg);
    ADispatchException(t);
    return AError;
}


AValue ARaiseValueError(AThread *t, const char *msg)
{
    ARaiseByNum(t, AErrorClassNum[EX_VALUE_ERROR], msg);
    ADispatchException(t);
    return AError;
}


AValue ARaiseIoError(AThread *t, const char *msg)
{
    return ARaiseByNum(t, AErrorClassNum[EX_IO_ERROR], msg);
}


AValue ARaiseRuntimeError(AThread *t, const char *msg)
{
    return ARaiseByNum(t, AErrorClassNum[EX_RUNTIME_ERROR], msg);
}


AValue ARaiseIndexError(AThread *t, const char *msg)
{
    return ARaiseByNum(t, AErrorClassNum[EX_INDEX_ERROR], msg);
}


AValue ARaiseErrnoException(AThread *t, const char *type)
{
    int num = AGetGlobalNum(t, type);
    return ARaiseErrnoExceptionByNum(t, num);
}


ABool AIsExceptionType(AThread *t, const char *type)
{
    AValue typeVal = AGlobal(t, type);
    return AIsValue(t, AGetException(t), typeVal);
}


int AMemberNum(AThread *t, const char *member)
{
    int id;
    ASymbol *sym;
    ASymbolInfo *symInfo;
    
    ALockInterpreter();
    if (!AGetSymbol(member, strlen(member), &sym))
        goto Fail;
    symInfo = AGetMemberSymbol(sym);
    if (symInfo == NULL)
        goto Fail;
    id = symInfo->num;
    AUnlockInterpreter();
    
    return id;

  Fail:
    
    AUnlockInterpreter();
    return ARaiseMemoryError(t); 
}


AValue AMember(AThread *t, AValue o, const char *member)
{
    int id;
    
    *t->tempStack = o;
    id = AMemberNum(t, member);
    o = *t->tempStack;
    *t->tempStack = AZero;
    
    return AMemberByNum(t, o, id);
}


AValue ASetMember(AThread *t, AValue o, const char *member, AValue v)
{
    int id;
    AValue ret;
    
    t->tempStack[0] = o;
    t->tempStack[1] = v;
    id = AMemberNum(t, member);
    
    ret = ASetMemberHelper(t, t->tempStack, id);
    t->tempStack[0] = t->tempStack[1] = AZero;
    return ret;
}


AValue ASuperMember(AThread *t, AValue *frame, const char *member)
{
    int id = AMemberNum(t, member);
    ATypeInfo *type = AGetInstanceType(AValueToInstance(frame[0]));

    while (type->super != NULL) {
        int method, read;
        
        type = type->super;

        method = ALookupMemberTable(type, MT_METHOD_PUBLIC, id);
        if (method != -1)
            return ACreateBoundMethod(t, frame[0], method);
        else {
            read = ALookupMemberTable(type, MT_VAR_GET_PUBLIC, id);
            if (read != -1) {
                /* Direct access or call getter? */
                if (read < A_VAR_METHOD)
                    return AMemberDirect(frame[0], read);
                else
                    return ACallValue(t, AGlobalByNum(read & ~A_VAR_METHOD), 1,
                                      frame);
            }
        }
    }
    
    return ARaiseMemberErrorND(t, frame[0], id);
}


AValue ASetSuperMember(AThread *t, AValue *frame, const char *member, AValue v)
{
    int id = AMemberNum(t, member);
    ATypeInfo *type = AGetInstanceType(AValueToInstance(frame[0]));

    while (type->super != NULL) {
        int write;
        
        type = type->super;

        write = ALookupMemberTable(type, MT_VAR_SET_PUBLIC, id);
        if (write != -1) {
            /* Direct access or call setter? */
            if (write < A_VAR_METHOD) {
                ASetMemberDirect(t, frame[0], write, v);
                return ANil;
            } else {
                AValue ret;
                t->tempStack[0] = frame[0];
                t->tempStack[1] = v;
                ret = ACallValue(t, AGlobalByNum(write & ~A_VAR_METHOD), 2,
                            t->tempStack);
                t->tempStack[0] = AZero;
                t->tempStack[1] = AZero;
                if (!AIsError(ret))
                    ret = ANil;
                return ret;
            }
        }
        
    }

    return ARaiseMemberErrorND(t, frame[0], id);
}


AValue ACall(AThread *t, const char *callable, int numArgs, AValue *args)
{
    AValue callValue = AGlobal(t, callable);
    return ACallValue(t, callValue, numArgs, args);
}


AValue ACallMethod(AThread *t, const char *member, int nargs, AValue *args)
{
    int id = AMemberNum(t, member);
    return ACallMethodByNum(t, id, nargs, args);
}


void ASetMemberDirect(AThread *t, AValue o, int num, AValue newVal)
{
    *t->tempStack = newVal;
    if (!ASetInstanceMember(t, AValueToInstance(o), num, t->tempStack))
        ARaiseMemoryError(t);
}


AValue ASetMemberByNum(AThread *t, AValue o, int member, AValue v)
{
    AValue ret;
    
    t->tempStack[0] = o;
    t->tempStack[1] = v;
    
    ret = ASetMemberHelper(t, t->tempStack, member);
    t->tempStack[0] = t->tempStack[1] = AZero;
    return ret;
}


ABool AHasMemberByNum(AValue o, int num)
{
    ATypeInfo *type = AGetInstanceType(AValueToInstance(o));
    do {
        if (ALookupMemberTable(type, MT_METHOD_PUBLIC, num) != -1 ||
            ALookupMemberTable(type, MT_VAR_GET_PUBLIC, num) != -1)
            return TRUE;
        type = type->super;
    } while (type != NULL);

    return FALSE;
}


int AGetGlobalNum(AThread *t, const char *name)
{
    char name2[256 + 1];
    AToken *tok;
    AToken *tokTmp;
    ASymbolInfo *sym;
    int num;
    
    strcpy(name2, name);
    strcat(name2, "\n");
    
    ALockInterpreter();

    tok = NULL;
    if (!ATokenize((unsigned char *)name2, (unsigned char *)name2 +
                   strlen(name2), &tok, &tokTmp, NULL)) {
        AUnlockInterpreter();
        ARaiseMemoryError(t);
        return 0; /* Unreached */
    }
    
    AParseGlobalVariable(tok, &sym, TRUE, TRUE);
    AFreeTokens(tok);

    if (sym->type == ID_ERR_PARSE || sym->num == 0) {
        AUnlockInterpreter();
        ARaiseValueError(t, "Invalid global variable");
        return -1;
    } else
        num = sym->num;
    
    AUnlockInterpreter();

    return num;
}


AValue AGlobal(AThread *t, const char *name)
{
    return AGlobalByNum(AGetGlobalNum(t, name));
}


void AGetPair(AThread *t, AValue pair, AValue *v1, AValue *v2)
{
    if (!AIsPair(pair))
        ARaiseTypeError(t, AMsgPairExpected);
    *v1 = AValueToMixedObject(pair)->data.pair.head;
    *v2 = AValueToMixedObject(pair)->data.pair.tail;
}


static AValue AMakeMixed(AThread *t, int type, AValue v1, AValue v2)
{
    AMixedObject *obj;
    
    t->tempStack[0] = v1;
    t->tempStack[1] = v2;

    obj = AAlloc(t, AGetBlockSize(sizeof(AMixedObject)));
    if (obj == NULL)
        return ARaiseMemoryError(t);

    AInitValueBlock(&obj->header, 3 * sizeof(AValue));
    obj->type = type;
    obj->data.pair.head = t->tempStack[0];
    obj->data.pair.tail = t->tempStack[1];

    return AMixedObjectToValue(obj);
}


AValue AMakePair(AThread *t, AValue v1, AValue v2)
{
    return AMakeMixed(t, A_PAIR_ID, v1, v2);
}


void AGetRange(AThread *t, AValue pair, AValue *lo, AValue *hi)
{
    if (!AIsRange(pair))
        ARaiseTypeError(t, AMsgRangeExpected);
    *lo = AValueToMixedObject(pair)->data.range.start;
    *hi = AValueToMixedObject(pair)->data.range.stop;
}


AValue AMakeRange(AThread *t, AValue lo, AValue hi)
{
    AExpectInt(t, lo);
    AExpectInt(t, hi);
    return AMakeMixed(t, A_RANGE_ID, lo, hi);
}


int AIsEq(AThread *t, AValue a, AValue b)
{
    AValue res;
    
    if (AIsShortInt(a) && AIsShortInt(b))
        return a == b;

    if (AIsStr(a) && AIsStr(b))
        return ACompareStrings(a, b) == 0;

    if (AIsConstant(a))
        return a == b;

    res = AIsEqual(t, a, b);
    if (AIsError(res)) {
        ADispatchException(t);
        return -1;
    } else
        return res != AZero;
}


int AIsLt(AThread *t, AValue a, AValue b)
{
    AValue res;
    
    if (AIsShortInt(a) && AIsShortInt(b))
        return (ASignedValue)a < (ASignedValue)b;
    
    res = ACompareOrder(t, a, b, OPER_LT);
    if (AIsError(res)) {
        ADispatchException(t);
        return -1;
    } else
        return res != AZero;
}


int AIsGt(AThread *t, AValue a, AValue b)
{
    AValue res;
    
    if (AIsShortInt(a) && AIsShortInt(b))
        return (ASignedValue)a > (ASignedValue)b;
    
    res = ACompareOrder(t, a, b, OPER_GT);
    if (AIsError(res)) {
        ADispatchException(t);
        return -1;
    } else
        return res != AZero;
}


Assize_t ALen(AThread *t, AValue s)
{
    if (AIsArray(s))
        return AArrayLen(s);
    else {
        AValue *arg = AAllocTemp(t, s);        
        AValue ret = ACallMethodByNum(t, AM_LENGTH, 0, arg);
        AFreeTemp(t);
        if (AIsError(ret))
            return -1;
        else
            return AGetInt(t, ret);
    }
}


AValue AGetItemAt(AThread *t, AValue s, Assize_t i)
{
    if (AIsArray(s) && i >= 0)
        return AArrayItem(s, i); /* IDEA: Check range error? */
    else {
        AValue *args = AAllocTemps(t, 2);
        AValue ret;
        args[0] = s;
        args[1] = AIntToValue(i);
        ret = ACallMethodByNum(t, AM_GET_ITEM, 1, args);
        AFreeTemps(t, 2);
        return ret;
    }
}


AValue AGetItem(AThread *t, AValue s, AValue index)
{
    if (AIsInstance(s)) {
        if (AIsArray(s) && AIsShortInt(index)) {
            ASignedValue i = AValueToInt(index);
            if (i < 0 || i >= AArrayLen(s)) {
                i = ANormIndex(i, AArrayLen(s));
                if (i < 0 || i >= AArrayLen(s))
                    return ARaiseByNum(t, AErrorClassNum[EX_INDEX_ERROR],
                                       NULL);
            }
            return AArrayItem(s, i);
        } else {
            AValue *args = AAllocTemps(t, 2);
            AValue ret;
            args[0] = s;
            args[1] = index;
            ret = ACallMethodByNum(t, AM_GET_ITEM, 1, args);
            AFreeTemps(t, 2);
            return ret;
        }
    } else if (AIsShortInt(index)) {
        ABool result;

        /* Different cases (narrow/wide) of getting a single character are
           implemented separately for better performance. */
        
        if (AIsNarrowStr(s) || AIsNarrowSubStr(s)) {
            ASignedValue i = AValueToInt(index);
            AString *str;
        
            if (i < 0 || i >= AStrLen(s)) {
                i = ANormIndex(i, AStrLen(s));
                if (i < 0 || i >= AStrLen(s))
                    return ARaiseByNum(t, AErrorClassNum[EX_INDEX_ERROR],
                                       NULL);
            }
            
            /* Create a short string object of length 1. */
            AAlloc_M(t, AGetBlockSize(sizeof(AValue) + 1), str, result);
            if (!result)
                return ARaiseMemoryError(t);
            AInitNonPointerBlock(&str->header,  1);
            str->elem[0] = AStrItem(s, i);
            
            return AStrToValue(str);
        } else if (AIsWideStr(s) || AIsSubStr(s)) {
            ASignedValue i = AValueToInt(index);
            AWideString *str;
            
            if (i < 0 || i >= AStrLen(s)) {
                i = ANormIndex(i, AStrLen(s));
                if (i < 0 || i >= AStrLen(s))
                    return ARaiseByNum(t, AErrorClassNum[EX_INDEX_ERROR],
                                       NULL);
            }

            /* Create a wide string object of length 1. */
            AAlloc_M(t, AGetBlockSize(sizeof(AValue) + sizeof(AWideChar)), str,
                    result);
            if (!result)
                return ARaiseMemoryError(t);            
            AInitNonPointerBlock(&str->header, sizeof(AWideChar));
            str->elem[0] = AStrItem(s, i);
            
            return AWideStrToValue(str);
        } else
            return ARaiseTypeError(t, NULL); /* IDEA: Provide error message. */
    } else if (AIsPair(index)) {
        AValue loVal, hiVal;
        Assize_t lo, hi;
        
        AGetPair(t, index, &loVal, &hiVal);

        if (AIsNil(loVal))
            loVal = AZero;
        if (AIsNil(hiVal))
            hiVal = AIntToValue(A_SLICE_END);
        
        lo = AGetInt_ssize_t(t, loVal);
        hi = AGetInt_ssize_t(t, hiVal);

        if (AIsStr(s))
            return ASubStr(t, s, lo, hi);
        else
            return ARaiseTypeError(t, NULL); /* IDEA: Provide error message. */
    } else
        return ARaiseTypeError(t, NULL); /* IDEA: Provide error message. If
                                            the index is long int, change
                                            exception! */
}


AValue ASetItemAt(AThread *t, AValue s, Assize_t i, AValue o)
{
    if (AIsArray(s) && i >= 0)
        ASetArrayItem(t, s, i, o); /* IDEA: Check range error? */
    else {
        AValue *args = AAllocTemps(t, 3);
        AValue ret;
        args[0] = s;
        args[1] = AIntToValue(i);
        args[2] = o;
        ret = ACallMethodByNum(t, AM_SET_ITEM, 2, args);
        AFreeTemps(t, 3);
        if (AIsError(ret))
            return ret;
    }
    return ANil;
}


AValue ASetItem(AThread *t, AValue s, AValue index, AValue o)
{
    if (AIsInstance(s)) {
        AValue *args;
        AValue ret;

        if (AIsArray(s) && AIsShortInt(index)) {
            Assize_t len = AArrayLen(s);
            ASignedValue i = AValueToInt(index);
            i = ANormIndex(i, len);
            if (i >= 0 && i < len) {
                ASetArrayItem(t, s, i, o);
                return ANil;
            }
        }
        
        args = AAllocTemps(t, 3);
        args[0] = s;
        args[1] = index;
        args[2] = o;
        ret = ACallMethodByNum(t, AM_SET_ITEM, 2, args);
        AFreeTemps(t, 3);
        if (AIsError(ret))
            return ret;
    } else
        return ARaiseTypeError(t, NULL); /* IDEA: Provide an error message. */
    
    return ANil;
}


static AValue InstanceOp(AThread *t, AValue a, AValue b, int member)
{
    AValue *args = AAllocTemps(t, 2);
    AValue ret;
    
    args[0] = a;
    args[1] = b;
    ret = ACallMethodByNum(t, member, 1, args);

    AFreeTemps(t, 2);

    return ret;
}


AValue AAdd(AThread *t, AValue a, AValue b)
{
    AValue sum;

    if (AIsShortInt(a) && AIsShortInt(b)) {
        sum = a + b;
        if (!AIsAddOverflow(sum, a, b))
            return sum;
        else 
            sum = ACreateLongIntFromIntND(t, AValueToInt(a) + AValueToInt(b));
    } else if (AIsInstance(a))
        return InstanceOp(t, a, b, AM_ADD);
    else if (AIsInstance(b) && (AIsInt(a) || AIsFloat(a)))
        return InstanceOp(t, b, a, AM_ADD);
    else if (AIsStr(a))
        return AConcat(t, a, b);
    else {
        a = ACoerce(t, a, b, (t->tempStack + 1));
        if (AIsError(a))
            sum = AError;
        else {        
            b = t->tempStack[1];
            
            if (AIsLongInt(a) && AIsLongInt(b))
                sum = AAddLongInt(t, a, b);
            else if (AIsFloat(a))
                return AMakeFloat(t, AValueToFloat(a) + AValueToFloat(b));
            else
                return ARaiseTypeError(t, NULL);
        }
    }

    if (AIsError(sum))
        ADispatchException(t);

    return sum;
}


AValue ASub(AThread *t, AValue a, AValue b)
{
    AValue diff;
    
    if (AIsShortInt(a) && AIsShortInt(b)) {
        diff = a - b;
        if (!AIsSubOverflow(diff, a, b))
            return diff;
        else
            diff = ACreateLongIntFromIntND(t, AValueToInt(a) - AValueToInt(b));
    } else if (AIsInstance(a))
        return InstanceOp(t, a, b, AM_SUB);
    else {
        a = ACoerce(t, a, b, (t->tempStack + 1));
        if (AIsError(a))
            diff = AError;
        else {
            b = t->tempStack[1];
            
            if (AIsLongInt(a) && AIsLongInt(b))
                diff = ASubLongInt(t, a, b);
            else if (AIsFloat(a))
                return AMakeFloat(t, AValueToFloat(a) - AValueToFloat(b));
            else
                return ARaiseTypeError(t, NULL);
        }
    }

    if (AIsError(diff))
        ADispatchException(t);

    return diff;
}

    
AValue AMul(AThread *t, AValue a, AValue b)
{
    AValue prod;
    
    if (AIsShortInt(a) && AIsShortInt(b)) {
        if (AHi(a) == 0 && AHi(b) == 0) {
            prod = AValueToInt(a) * (ASignedValue)b;
            if ((ASignedValue)prod >= 0)
                return prod;
        }

        prod = AMultiplyShortInt(t, a, b);
    } else if (AIsInstance(a))
        return InstanceOp(t, a, b, AM_MUL);
    else if (AIsInstance(b) && (AIsInt(a) || AIsFloat(a)))
        return InstanceOp(t, b, a, AM_MUL);
    else if (AIsShortInt(a) && AIsStr(b))
        return ARepeatString(t, b, AValueToInt(a));
    else if (AIsShortInt(b) && AIsStr(a))
        return ARepeatString(t, a, AValueToInt(b));
    else {
        a = ACoerce(t, a, b, (t->tempStack + 1));
        if (AIsError(a))
            prod = AError;
        else {
            b = t->tempStack[1];
            
            if (AIsLongInt(a) && AIsLongInt(b))
                prod = AMultiplyLongInt(t, a, b);
            else if (AIsFloat(a))
                return AMakeFloat(t, AValueToFloat(a) * AValueToFloat(b));
            else
                return ARaiseTypeError(t, NULL);
        }
    }

    if (AIsError(prod))
        ADispatchException(t);

    return prod;
}


static AValue DivModByZero(AThread *t)
{
    ARaiseByNum(t, AErrorClassNum[EX_ARITHMETIC_ERROR], AMsgDivisionByZero);
    ADispatchException(t);
    return AError;
}


AValue ADiv(AThread *t, AValue a, AValue b)
{
    if (b == AZero)
        return DivModByZero(t);
    
    if (AIsShortInt(a) && AIsShortInt(b))
        return AMakeFloat(t, (double)AValueToInt(a) / AValueToInt(b));
    
    if (AIsInstance(a))
        return InstanceOp(t, a, b, AM_DIV);

    a = ACoerce(t, a, b, t->tempStack + 1);
    if (AIsError(a)) {
        ADispatchException(t);
        return AError;
    }
    
    b = t->tempStack[1];
    
    if (AIsLongInt(a) && AIsLongInt(b))
        return AMakeFloat(t, ALongIntToFloat(a) / ALongIntToFloat(b));
    else if (AIsFloat(a)) {
        if (AValueToFloat(b) == 0.0)
            return DivModByZero(t);
        return AMakeFloat(t, AValueToFloat(a) / AValueToFloat(b));
    } else
        return ARaiseTypeError(t, NULL);
}


AValue AIntDiv(AThread *t, AValue a, AValue b)
{
    AValue mod;
    AValue div;
    
    if (AIsShortInt(a) && AIsShortInt(b))
        div = AIntDivMod(t, a, b, &mod);
    else if (AIsInstance(a))
        return InstanceOp(t, a, b, AM_IDIV);
    else if (b == AZero)
        return DivModByZero(t);
    else {
        a = ACoerce(t, a, b, t->tempStack + 1);
        if (AIsError(a))
            div = AError;
        else {
            b = t->tempStack[1];
            
            if (AIsLongInt(a) && AIsLongInt(b))
                div = ADivModLongInt(t, a, b, t->tempStack + 1);
            else if (AIsFloat(a)) {
                if (AValueToFloat(b) == 0.0)
                    return DivModByZero(t);
                return AMakeFloat(t, floor(AValueToFloat(a) /
                                           AValueToFloat(b)));
            } else
                return ARaiseTypeError(t, NULL);
        }
    }

    if (AIsError(div))
        ADispatchException(t);

    return div;
}


AValue AMod(AThread *t, AValue a, AValue b)
{
    AValue mod;
    
    if (AIsShortInt(a) && AIsShortInt(b)) {
        if (AIntDivMod(t, a, b, &mod) == AError)
            ADispatchException(t);
        return mod;
    }
    
    if (AIsInstance(a))
        return InstanceOp(t, a, b, AM_MOD);

    if (b == AZero)
        return DivModByZero(t);

    a = ACoerce(t, a, b, t->tempStack + 1);
    if (AIsError(a))
        ADispatchException(t);
    
    b = t->tempStack[1];
    
    if (AIsLongInt(a) && AIsLongInt(b)) {
        if (ADivModLongInt(t, a, b, &mod) == AError)
            ADispatchException(t);
        return mod;
    } else if (AIsFloat(a)) {
        double result;
        double bFloat = AValueToFloat(b);
         
        if (bFloat == 0.0)
            return DivModByZero(t);
        result = fmod(AValueToFloat(a), bFloat);
        if (result * bFloat < 0)
            result += bFloat;
        return AMakeFloat(t, result);
    } else
        return ARaiseTypeError(t, NULL);
}


AValue APow(AThread *t, AValue a, AValue b)
{
    if (AIsInt(a) && AIsInt(b)) {
        AValue p = APowInt(t, a, b);
        if (!AIsError(p))
            return p;
        else
            ADispatchException(t);
    }
    
    if (AIsInstance(a))
        return InstanceOp(t, a, b, AM_POW);

    a = ACoerce(t, a, b, t->tempStack + 1);
    if (AIsError(a))
        ADispatchException(t);
    
    b = t->tempStack[1];
    
    if (AIsFloat(a)) {
        if (AValueToFloat(a) < 0.0) {
            ARaiseByNum(t, AErrorClassNum[EX_ARITHMETIC_ERROR], NULL);
            ADispatchException(t);
            return AError;
        }
        return AMakeFloat(t, pow(AValueToFloat(a), AValueToFloat(b)));
    } else
        return ARaiseTypeError(t, NULL);
}


AValue ANeg(AThread *t, AValue a)
{
    AValue neg;
    
    if (AIsShortInt(a)) {
        if (a == AIntToValue(A_SHORT_INT_MIN))
            neg = ACreateLongIntFromIntND(t, -A_SHORT_INT_MIN);
        else
            return -(ASignedValue)a;
    } else if (AIsFloat(a))
        return AMakeFloat(t, -AValueToFloat(a));
    else if (AIsLongInt(a))
        neg = ANegateLongInt(t, a);
    else if (AIsInstance(a)) {
        AValue *arg = AAllocTemp(t, a);
        neg = ACallMethodByNum(t, AM_NEGATE, 0, arg);
        AFreeTemp(t);
        return neg; /* If a is an instance, do not dispatch exceptions. */
    } else
        return ARaiseTypeError(t, NULL);

    if (AIsError(neg))
        ADispatchException(t);

    return neg;
}


AValue AIterator(AThread *t, AValue v)
{
    AValue *args;
    AValue ret;

    args = AAllocTemps(t, 2);
    args[0] = v;
    ret = AStdIter(t, args);
    AFreeTemps(t, 2);
    
    return ret;
}


int ANext(AThread *t, AValue e, AValue *next)
{
    AValue *args;
    AValue ret;
    int status;

    args = AAllocTemps(t, 2);
    args[0] = e;
    ret = ACallMethod(t, "hasNext", 0, args);
    if (AIsError(ret))
        status = -1;
    else if (AIsTrue(ret)) {
        ret = ACallMethod(t, "next", 0, args);
        if (AIsError(ret))
            status = -1;
        else {
            *next = ret;
            status = 1;
        }
    } else if (AIsFalse(ret))
        status = 0;
    else {
        ARaiseValueError(t, "Invalid boolean value");
        status = -1;
    }
    AFreeTemps(t, 2);

    return status;
}


int AIn(AThread *t, AValue a, AValue b)
{
    AValue res = AIsIn(t, a, b);
    if (AIsError(res))
        return -1;
    else if (res == AZero)
        return 0;
    else
        return 1;
}


AValue AMakeUninitializedObject(AThread *t, AValue type)
{
    AInstance *inst;
    
    inst = AAlloc(t, AValueToType(type)->instanceSize);
    if (inst == NULL)
        return ARaiseMemoryError(t);
    
    AInitInstanceBlock(&inst->type, AValueToType(type));
    AInitInstanceData(inst, AValueToType(type));
    
    return AInstanceToValue(inst);
}


ABool AIsStandalone(void)
{
    return AIsStandaloneFlag;
}


AValue AAllocContainer(AThread *t, AValue initValue)
{
    AValue *tmp;
    AFixArray *container;
    AValue ret;

    tmp = AAllocTemps(t, 2);
    
    /* NOTE: The container is always allocated in the old generation. */
    
    tmp[0] = initValue;

    /* Allocate and initialize a container object, implemented as a FixArray
       object. */
    container = AAllocUnmovable(2 * sizeof(AValue));
    AInitValueBlockOld(&container->header, sizeof(AValue));
    container->elem[0] = AZero;

    /* Store the initial value. */
    tmp[1] = AFixArrayToValue(container);    
    ASetFixArrayItem(t, tmp[1], 0, tmp[0]);
    
    ret = tmp[1];
    
    AFreeTemps(t, 2);
    
    return ret;
}


void *AContainerPtr(AValue value)
{
    return AValueToFixArray(value);
}


AValue AContainerValue(void *ptr)
{
    AFixArray *container = ptr;
    return container->elem[0];
}


void ASetContainerValue(AThread *t, void *ptr, AValue item)
{
    AValue val = AFixArrayToValue((AFixArray *)ptr);
    ASetFixArrayItem(t, val, 0, item);
}


/* Write Repr(val) to stdout followed by a line break. */
AValue AWriteRepr(AThread *t, AValue val)
{
    AValue *temp = AAllocTemps(t, 2);

    temp[0] = val;
    temp[0] = ACall(t, "std::Repr", 1, temp);
    if (AIsError(temp[0]))
        return AError;
    return ACall(t, "std::WriteLn", 1, temp);
}


AValue AEpicInternalFailure(const char *message)
{
    fprintf(stderr, "\n*** UNRECOVERABLE INTERNAL ALORE ERROR ***\n"
            "Reason: %s\nPlease file a bug report!\n", message);
    exit(99);
}


AValue AHash(AThread *t, AValue object)
{
    AValue hash;
    AValue *frame = AAllocTemps(t, 3);

    frame[0] = object;
    hash = AStdHash(t, frame);    
    
    AFreeTemps(t, 3);

    return hash;
}
