/* std_str.c - std::Str related operations 

   Copyright (c) 2010-2011 Jukka Lehtosalo

   Alore is licensed under the terms of the MIT license.  See the file
   LICENSE.txt in the distribution.
*/

#include "alore.h"
#include "runtime.h"
#include "memberid.h"
#include "mem.h"
#include "str.h"
#include "int.h"
#include "str_internal.h"
#include "internal.h"
#include "debug_runtime.h"
#include "wrappers.h"
#include "gc.h"
#include "util.h"
#include "symtable.h"
#include "std_module.h"


#define INT_BUF_SIZE (1 + ((A_VALUE_INT_BITS - 1) * 4 + 12) / 13)


int AStrIterNum;


static AValue SymbolStr(AThread *t, ASymbolInfo *sym);
static ABool IsMatchAt(AValue src, Assize_t i, AValue dst, Assize_t dstLen);
static Assize_t Find(AValue str, Assize_t i, AValue substr);
static Assize_t Count(AValue str, AValue substr, Assize_t max);


#define STR_ITER_I 0
#define STR_ITER_LEN 1
#define STR_ITER_S 2


/* NOTE: frame may point almost anywhere! Be careful. */
/* Always returns either a string value or exception. Frame contains exactly
   one value. */
AValue AStdStr(AThread *t, AValue *frame)
{
    /* IDEA: Implement repetive type checking "if" checks using a switch
             statement -- it might be a lot faster in some cases. */
    
    if (AIsShortInt(frame[0])) {
        unsigned char buf[INT_BUF_SIZE];
        Assize_t ind;
        ASignedValue n;
        ABool isNeg;
        AString *s;
        Assize_t len;
        ABool result;
        unsigned char *dst;

        ind = INT_BUF_SIZE - 1;
        
        n = AValueToInt(frame[0]);
        if (n < 0) {
            n = -n;
            isNeg = TRUE;
        } else
            isNeg = FALSE;
        
        while (n >= 10) {
            buf[ind--] = n % 10 + '0';
            n /= 10;
        }
        
        buf[ind] = n + '0';

        if (isNeg)
            buf[--ind] = '-';

        len = INT_BUF_SIZE - ind;

        AAlloc_M(t, AGetBlockSize(sizeof(AValue) + len), s, result);
        if (!result)
            return ARaiseMemoryErrorND(t);

        AInitNonPointerBlock(&s->header, len);
        
        dst = s->elem;
        
        do {
            *dst++ = buf[ind++];
        } while (ind < INT_BUF_SIZE);

        return AStrToValue(s);
    } else if (AIsStr(frame[0]))
        return frame[0];
    else if (AIsInstance(frame[0])) {
        AValue v = ACallMethodByNum(t, AM__STR, 0, frame);
        /* Expect a Str object. */
        if (AIsStr(v))
            return v;
        else if (AIsError(v)) {
            if (AMemberByNum(t, frame[0], AM__STR) == AError) {
                char buf[500];
                AFormatMessage(buf, sizeof(buf), "<%q instance>",
                               AGetInstanceType(
                                   AValueToInstance(frame[0]))->sym);
                v = ACreateStringFromCStr(t, buf);
            }
            return v;
        } else
            return ARaiseTypeErrorND(t);
    } else if (AIsFloat(frame[0])) {
        /* IDEA: sprintf might produce inconsistent results compared with
                 format() of Str. Verify that they are reasonable. */
        /* IDEA: Verify that the length of the buffer is reasonable. */
        char buf[100];
        sprintf(buf, "%.10g", AValueToFloat(frame[0]));
        /* If the float is an infinity or a NaN, the result of sprintf varies
           from system to system. Normalize any non-standard values. */
        AFixInfAndNanStrings(buf);
        return ACreateStringFromCStr(t, buf);
    } else if (AIsLongInt(frame[0]))
        return ALongIntToStr(t, frame[0], 10, 0);
    else if (AIsGlobalFunction(frame[0]))
        return SymbolStr(t, AValueToFunction(frame[0])->sym);
    else if (AIsNonSpecialType(frame[0]))
        return SymbolStr(t, AValueToType(frame[0])->sym);
    else if (AIsConstant(frame[0])) {
        return SymbolStr(t, AValueToConstant(frame[0])->sym);
    } else
        return AStdRepr(t, frame);
}


AValue AStrLengthMethod(AThread *t, AValue *frame)
{
    frame[0] = AWrapObject(t, frame[0], AM_LENGTH);
    return AMakeInt_ssize_t(t, AStrLen(A_UNWRAP(frame[0])));
}


AValue AStrCount(AThread *t, AValue *frame)
{
    Assize_t n;

    frame[0] = AWrapObject(t, frame[0], -1);
    AExpectStr(t, frame[1]);
    frame[2] = A_UNWRAP(frame[0]);

    n = Count(frame[2], frame[1], A_SSIZE_T_MAX);
    return AMakeInt_ssize_t(t, n);
}


AValue AStrUpper(AThread *t, AValue *frame)
{
    Assize_t begIndex;
    Assize_t len;

    frame[0] = AWrapObject(t, frame[0], -1);
    frame[1] = A_UNWRAP(frame[0]);

    if (AIsNarrowStr(frame[1])) {
        AString *s;
        unsigned char *src;
        Assize_t i;
        
        begIndex = 0;
        len = AGetStrLen(AValueToStr(frame[1]));

      Narrow:

        /* FIX: perhaps inline.. */
        s = AAlloc(t, sizeof(AValue) + len);
        if (s == NULL)
            return AError;
        
        AInitNonPointerBlock(&s->header, len);

        src = AValueToStr(frame[1])->elem + begIndex;

        for (i = 0; i < len; i++) {
            if (src[i] == 255)
                goto ConvertToWide;
            s->elem[i] = AUpperCase[src[i]];
        }

        return AStrToValue(s);

      ConvertToWide:

        frame[1] = ANarrowStringToWide(t, frame[0], frame);
        if (AIsError(frame[1]))
            return AError;
        goto Wide;      
    } else if (AIsWideStr(frame[1])) {
        AWideString *s;
        AWideChar *src;
        Assize_t i;
        
        begIndex = 0;
        len = AGetWideStrLen(AValueToWideStr(frame[1]));

      Wide:

        s = AAlloc(t, sizeof(AValue) + sizeof(AWideChar) * len);
        if (s == NULL)
            return AError;
        
        AInitNonPointerBlock(&s->header, sizeof(AWideChar) * len);
        
        src = AValueToWideStr(frame[1])->elem + begIndex;

        for (i = 0; i < len; i++) {
            unsigned ch = src[i];
            s->elem[i] = AUpper(ch);
        }

        return AWideStrToValue(s);
    } else if (AIsSubStr(frame[1])) {
        ASubString *ss = AValueToSubStr(frame[1]);

        begIndex = AValueToInt(ss->ind);
        len = AValueToInt(ss->len);

        frame[1] = ss->str;
        if (AIsNarrowStr(frame[1]))
            goto Narrow;
        else
            goto Wide;
    } else
        return ARaiseTypeErrorND(t);

    /* Not reached */
}


AValue AStrLower(AThread *t, AValue *frame)
{
    Assize_t begIndex;
    Assize_t len;    

    frame[0] = AWrapObject(t, frame[0], -1);
    frame[1] = A_UNWRAP(frame[0]);
    
    if (AIsNarrowStr(frame[1])) {
        AString *s;
        unsigned char *src;
        Assize_t i;
        
        begIndex = 0;
        len = AGetStrLen(AValueToStr(frame[1]));

      Narrow:

        s = AAlloc(t, sizeof(AValue) + len);
        if (s == NULL)
            return AError;
        
        AInitNonPointerBlock(&s->header, len);

        src = AValueToStr(frame[1])->elem + begIndex;

        for (i = 0; i < len; i++)
            s->elem[i] = ALowerCase[src[i]];

        return AStrToValue(s);
    } else if (AIsWideStr(frame[1])) {
        AWideString *s;
        AWideChar *src;
        Assize_t i;
        
        begIndex = 0;
        len = AGetWideStrLen(AValueToWideStr(frame[1]));

      Wide:

        s = AAlloc(t, sizeof(AValue) + sizeof(AWideChar) * len);
        if (s == NULL)
            return AError;
        
        AInitNonPointerBlock(&s->header, sizeof(AWideChar) * len);
        
        src = AValueToWideStr(frame[1])->elem + begIndex;

        for (i = 0; i < len; i++) {
            unsigned ch = src[i];
            s->elem[i] = ALower(ch);
        }

        return AWideStrToValue(s);
    } else if (AIsSubStr(frame[1])) {
        ASubString *ss = AValueToSubStr(frame[1]);

        begIndex = AValueToInt(ss->ind);
        len = AValueToInt(ss->len);

        frame[1] = ss->str;
        if (AIsNarrowStr(frame[1]))
            goto Narrow;
        else
            goto Wide;
    } else
        return ARaiseTypeErrorND(t);

    /* Not reached */
}


AValue AStrStrip(AThread *t, AValue *frame)
{
    Assize_t i1, i2, len;
    AValue str;
    
    frame[0] = AWrapObject(t, frame[0], -1);
    str = A_UNWRAP(frame[0]);

    len = AStrLen(str);
    i1 = 0;
    while (i1 < len) {
        int ch = AStrItem(str, i1);
        if (ch != ' ' && ch != '\t' && ch != '\n' && ch != '\r')
            break;
        i1++;
    }
    
    i2 = len - 1;
    while (i2 > i1) {
        int ch = AStrItem(str, i2);
        if (ch != ' ' && ch != '\t' && ch != '\n' && ch != '\r')
            break;
        i2--;
    }

    return ASubStr(t, str, i1, i2 + 1);
}


AValue AStrFind(AThread *t, AValue *frame)
{
    Assize_t index;
    Assize_t startIndex;
    
    frame[0] = AWrapObject(t, frame[0], -1);
    
    AExpectStr(t, frame[1]);

    if (!AIsDefault(frame[2])) {
        startIndex = AGetInt(t, frame[2]);
        if (startIndex < 0)
            return ARaiseValueError(t, "Negative start index");
    } else
        startIndex = 0;

    index = Find(A_UNWRAP(frame[0]), startIndex, frame[1]);

    if (index >= 0)
        return AMakeInt_ssize_t(t, index);
    else
        return AMakeInt(t, -1);
}


AValue AStrIndex(AThread *t, AValue *frame)
{
    Assize_t index;
    
    frame[0] = AWrapObject(t, frame[0], -1);
    
    AExpectStr(t, frame[1]);

    index = Find(A_UNWRAP(frame[0]), 0, frame[1]);

    if (index >= 0)
        return AMakeInt_ssize_t(t, index);
    else
        return ARaiseValueError(t, "Substring not found");
}


AValue AStrFormat(AThread *t, AValue *frame)
{
    frame[0] = AWrapObject(t, frame[0], -1);
    frame[2] = frame[1];
    frame[1] = A_UNWRAP(frame[0]);
    return AFormat(t, frame + 1);
}


AValue AStrStartsWith(AThread *t, AValue *frame)
{
    Assize_t len, prefixLen, i;
    AValue str;
    
    frame[0] = AWrapObject(t, frame[0], -1);
    str = A_UNWRAP(frame[0]);
    AExpectStr(t, frame[1]);

    len = AStrLen(str);
    prefixLen = AStrLen(frame[1]);

    if (len < prefixLen)
        return AFalse;

    for (i = 0; i < prefixLen; i++) {
        if (AStrItem(frame[1], i) != AStrItem(str, i))
            return AFalse;
    }
    
    return ATrue;
}


AValue AStrEndsWith(AThread *t, AValue *frame)
{
    Assize_t len, suffixLen, i;
    AValue str;
    
    frame[0] = AWrapObject(t, frame[0], -1);
    str = A_UNWRAP(frame[0]);
    AExpectStr(t, frame[1]);

    len = AStrLen(str);
    suffixLen = AStrLen(frame[1]);

    if (len < suffixLen)
        return AFalse;

    for (i = 1; i <= suffixLen; i++) {
        if (AStrItem(frame[1], suffixLen - i) != AStrItem(str, len - i))
            return AFalse;
    }
    
    return ATrue;
}


/* Convert a wide string to a narrow string. Assume wide to be either a wide
   string or a wide substring. Raise exception if unsuccessful. */
AValue AWideStringToNarrow(AThread *t, AValue wide, AValue *keep)
{
    Assize_t i;
    AWideChar *str;
    Assize_t len;
    AString *narrowStr;
    
    if (AIsSubStr(wide)) {
        str = AGetWideSubStrElem(wide);
        len = AGetSubStrLen(AValueToSubStr(wide));
    } else {
        str = AGetWideStrElem(wide);
        len = AGetWideStrLen(AValueToWideStr(wide));
    }
    
    for (i = 0; i < len; i++) {
        if (str[i] > 0xff)
            return ARaiseValueErrorND(t);
    }

    t->tempStack[0] = wide;
    if (keep != NULL)
        t->tempStack[1] = *keep;
    narrowStr = AAlloc(t, sizeof(AValue) + len);
    wide  = t->tempStack[0];
    if (keep != NULL)
        *keep = t->tempStack[1];

    if (narrowStr == NULL)
        return AError;
    
    AInitNonPointerBlock(&narrowStr->header, len);
    
    if (AIsSubStr(wide))
        str = AGetWideSubStrElem(wide);
    else
        str = AGetWideStrElem(wide);

    for (i = 0; i < len; i++)
        narrowStr->elem[i] = str[i];

    return AStrToValue(narrowStr);
}


/* Convert a narrow string to a wide string. Assume narrow to be a narrow
   string or a substring FIX???. May return a substring. FIX??? */
AValue ANarrowStringToWide(AThread *t, AValue narrow, AValue *keep)
{
    Assize_t i;
    unsigned char *str;
    Assize_t len;
    Assize_t blockSize;
    AWideString *wideStr;
    
    if (AIsSubStr(narrow)) {
        if (AIsWideStr(AValueToSubStr(narrow)->str))
            return narrow;
        len = AGetSubStrLen(AValueToSubStr(narrow));
    } else
        len = AGetStrLen(AValueToStr(narrow));

    /* FIX: Perhaps not like this.. call AllocKeep maybe??? */
    blockSize = AGetBlockSize(sizeof(AValue) + len * sizeof(AWideChar));
    if (t->heapPtr + blockSize > t->heapEnd || A_NO_INLINE_ALLOC) {
        t->tempStack[0] = narrow;
        t->tempStack[1] = *keep;
        
        wideStr = AAlloc(t, blockSize);
        
        narrow = t->tempStack[0];
        *keep  = t->tempStack[1];

        if (wideStr == NULL)
            return AError;
    } else {
        wideStr = (AWideString *)t->heapPtr;
        t->heapPtr += blockSize;
    }

    if (AIsSubStr(narrow))
        str = AGetSubStrElem(narrow);
    else
        str = AGetStrElem(narrow);

    AInitNonPointerBlock(&wideStr->header, len * sizeof(AWideChar));
    
    for (i = 0; i < len; i++)
        wideStr->elem[i] = str[i];

    return AWideStrToValue(wideStr);
}


ABool AIsNarrowStringContents(AValue str)
{
    if (AIsNarrowStr(str) || AIsNarrowSubStr(str))
        return TRUE;
    else {
        Assize_t i;
        Assize_t len;
        AWideChar *elem;

        len = AStrLen(str);
        elem = AWideStrItems(str);

        for (i = 0; i < len; i++) {
            if (elem[i] > 0xff)
                return FALSE;
        }

        return TRUE;
    }
}


/* Return the concatenation of two strings as a wide string. The arguments
   can be any string values. */
AValue AConcatWideStrings(AThread *t, AValue left, AValue right)
{
    Assize_t leftLen;
    Assize_t rightLen;
    Assize_t blockSize;
    AWideString *wideStr;
    AWideChar *leftStr;
    AWideChar *rightStr;

    if (AIsNarrowStr(left) || AIsNarrowSubStr(left))
        left = ANarrowStringToWide(t, left, &right);
    
    if (!AIsWideStr(right)) {
        if (AIsNarrowStr(right) || AIsNarrowSubStr(right))
            right = ANarrowStringToWide(t, right, &left);
        else if (!AIsSubStr(right))
            return ARaiseTypeErrorND(t);
    }

    if (AIsSubStr(left))
        leftLen = AGetSubStrLen(AValueToSubStr(left));
    else if (AIsWideStr(left))
        leftLen = AGetWideStrLen(AValueToWideStr(left));
    else
        return ARaiseTypeErrorND(t);

    if (AIsSubStr(right))
        rightLen = AGetSubStrLen(AValueToSubStr(right));
    else
        rightLen = AGetWideStrLen(AValueToWideStr(right));

    blockSize = AGetBlockSize(sizeof(AValue) + sizeof(AWideChar) *
                              (leftLen + rightLen));
    if (t->heapPtr + blockSize > t->heapEnd || A_NO_INLINE_ALLOC) {
        t->tempStack[0] = left;
        t->tempStack[1] = right;
        
        wideStr = AAlloc(t, blockSize);
        if (wideStr == NULL)
            return AError;
        
        left  = t->tempStack[0];
        right = t->tempStack[1];
    } else {
        wideStr = (AWideString *)t->heapPtr;
        t->heapPtr += blockSize;
    }

    if (AIsSubStr(left))
        leftStr = AGetWideSubStrElem(left);
    else
        leftStr = AGetWideStrElem(left);

    if (AIsSubStr(right))
        rightStr = AGetWideSubStrElem(right);
    else
        rightStr = AGetWideStrElem(right);

    AInitNonPointerBlock(&wideStr->header,
                         (leftLen + rightLen) * sizeof(AWideChar));

    ACopyMem(wideStr->elem, leftStr, leftLen * sizeof(AWideChar));
    ACopyMem(wideStr->elem + leftLen, rightStr, rightLen * sizeof(AWideChar));

    return AWideStrToValue(wideStr);
}


/* Convert any string to a narrow string. The returned string is never a
   substring object. If the conversion is unsuccessful, raise exception. */
AValue ANormalizeNarrowString(AThread *t, AValue strVal)
{
    if (!AIsNarrowStr(strVal)) {
        if (AIsWideStr(strVal)) {        
            strVal = AWideStringToNarrow(t, strVal, &strVal);
            if (AIsError(strVal))
                return AError;
        } else if (AIsSubStr(strVal)) {
            ASubString *ss;
            
            ss = AValueToSubStr(strVal);
            if (AIsWideStr(ss->str)) {
                strVal = AWideStringToNarrow(t, strVal, &strVal);
                if (AIsError(strVal))
                    return AError;
            } else {            
                AString *dst;
                AString *src;
                Assize_t len;
                Assize_t ind;

                len = AValueToInt(ss->len);
                ind = AValueToInt(ss->ind);

                *t->tempStack = ss->str;
                
                dst = AAlloc(t, sizeof(AValue) + len);
                if (dst == NULL)
                    return AError;
                
                src = AValueToStr(*t->tempStack);

                AInitNonPointerBlock(&dst->header, len);
                ACopyMem(dst->elem, src->elem + ind, len);

                return AStrToValue(dst);
            }
        } else 
            return ARaiseTypeErrorND(t);
    }

    return strVal;
}


/* Return the concatenation of two strings. */
AValue AConcatStrings(AThread *t, AValue left, AValue right)
{
    unsigned char *leftStr;
    unsigned char *rightStr;
    Assize_t leftLen;
    Assize_t rightLen;
    
    if (AIsNarrowStr(left)) {
        AString *res;
        Assize_t blockSize;
        Assize_t resultLen;
        
        leftLen = AGetStrLen(AValueToStr(left));
        leftStr = AValueToStr(left)->elem;
        
      NarrowLeft:
        
        if (AIsNarrowStr(right)) {
            rightLen = AGetStrLen(AValueToStr(right));
            rightStr = AValueToStr(right)->elem;
        } else if (AIsSubStr(right)) {
            ASubString *subStr;
            
            subStr = AValueToSubStr(right);
            if (AIsWideStr(subStr->str))
                return AConcatWideStrings(t, left, right);
            
            rightStr = AValueToStr(subStr->str)->elem +
                AValueToInt(subStr->ind);
            rightLen = AValueToInt(subStr->len);
        } else
            return AConcatWideStrings(t, left, right);
        
        resultLen = leftLen + rightLen;
        blockSize = AGetBlockSize(sizeof(AValue) + resultLen);
        
        if (t->heapPtr + blockSize > t->heapEnd || A_NO_INLINE_ALLOC) {
            t->tempStack[0] = left;
            t->tempStack[1] = right;
            
            res = AAlloc(t, blockSize);
            if (res == NULL)
                return AError;
            
            left = t->tempStack[0];
            if (AIsNarrowStr(left))
                leftStr = AGetStrElem(left);
            else
                leftStr = AGetSubStrElem(left);
            
            right = t->tempStack[1];
            if (AIsNarrowStr(right))
                rightStr = AGetStrElem(right);
            else
                rightStr = AGetSubStrElem(right);
        } else {
            res = (AString *)t->heapPtr;
            t->heapPtr += blockSize;
        }
        
        AInitNonPointerBlock(&res->header, resultLen);
        
        ACopyMem(res->elem, leftStr, leftLen);
        ACopyMem(res->elem + leftLen, rightStr, rightLen);
        
        return AStrToValue(res);
    }   
    
    if (AIsSubStr(left)) {
        ASubString *subStr;
        
        subStr = AValueToSubStr(left);
        
        if (AIsWideStr(subStr->str))
            return AConcatWideStrings(t, left, right);
        
        leftStr = AValueToStr(subStr->str)->elem +
            AValueToInt(subStr->ind);
        leftLen = AValueToInt(subStr->len);
        
        goto NarrowLeft;
    }
    
    if (AIsWideStr(left))
        return AConcatWideStrings(t, left, right);

    return ARaiseTypeErrorND(t);
}


#if A_MIN_SUBSTR_LEN != 5 && A_MIN_SUBSTR_LEN != 9
/* MIN_SUBSTR_LEN must be either 5 or 9 */
error
#endif


/* Create a substring of a string. Indices are 0-based. The index range is
   open at the end. */
AValue ACreateSubStr(AThread *t, AValue strVal, Assize_t begInd,
                     Assize_t endInd)
{
    Assize_t len;
    AString *str;
    AWideString *wideStr;
    Assize_t sLen;

    if (begInd < 0) {
        begInd = AStrLen(strVal) + begInd;
        if (begInd < 0)
            begInd = 0;
    }

    if (endInd < 0)
        endInd = AStrLen(strVal) + endInd;
     else if (endInd > AStrLen(strVal))
         endInd = AStrLen(strVal);

    if (endInd < begInd) {
        begInd = 0;
        endInd = 0;
    }

    len = endInd - begInd;

    if (AIsNarrowStr(strVal)) {
        AString *newStr;
        
        str = AValueToStr(strVal);
        sLen = AGetStrLen(str);

      NarrowStr:
        
        if (len < A_MIN_SUBSTR_LEN) {
            if (t->heapPtr + A_COPY_SUBSTR_SIZE > t->heapEnd
                || A_NO_INLINE_ALLOC) {
                *t->tempStack = strVal;
                newStr = AAlloc(t, A_COPY_SUBSTR_SIZE);
                if (newStr == NULL)
                    return AError;             
                strVal = *t->tempStack;
                str = AValueToStr(strVal);
            } else {
                newStr = (AString *)t->heapPtr;
                t->heapPtr += A_COPY_SUBSTR_SIZE;
            }

            AInitNonPointerBlock(&newStr->header, len);

            newStr->elem[0] = str->elem[begInd];
            newStr->elem[1] = str->elem[begInd + 1];
            newStr->elem[2] = str->elem[begInd + 2];
            newStr->elem[3] = str->elem[begInd + 3];
#if A_MIN_SUBSTR_LEN == 9
            if (len > 4) {
                newStr->elem[4] = str->elem[begInd + 4];
                newStr->elem[5] = str->elem[begInd + 5];
                newStr->elem[6] = str->elem[begInd + 6];
                newStr->elem[7] = str->elem[begInd + 7];
            }
#endif
            return AStrToValue(newStr);
        } else {
            /* Create a substr object */
            
            ASubString *ss;

          SubStr:
            
            if ((int)len < -1)
                goto IndexError;

            if (t->heapPtr + A_SUBSTR_SIZE > t->heapEnd
                || A_NO_INLINE_ALLOC) {
                *t->tempStack = strVal;
                ss = AAlloc(t, A_SUBSTR_SIZE);
                if (ss == NULL)
                    return AError;
                strVal = *t->tempStack;
            } else {
                ss = (ASubString *)t->heapPtr;
                t->heapPtr += A_SUBSTR_SIZE;
            }

            AInitValueBlock(&ss->header, sizeof(ASubString) - sizeof(AValue));

            ss->str = strVal;
            ss->ind = AIntToValue(begInd);
            ss->len = AIntToValue(len);

            return ASubStrToValue(ss);
        }
    } else if (AIsSubStr(strVal)) {
        ASubString *ss;

        ss = AValueToSubStr(strVal);

        sLen = AValueToInt(ss->ind) + AValueToInt(ss->len);
        
        begInd += AValueToInt(ss->ind);
        endInd += AValueToInt(ss->ind);

        strVal = ss->str;

        if (AIsNarrowStr(strVal)) {
            str = AValueToStr(strVal);
            goto NarrowStr;
        } else {
            wideStr = AValueToWideStr(strVal);
            goto WideStr;
        }
    } else if (AIsWideStr(strVal)) {
        AWideString *newWideStr;
        
        wideStr = AValueToWideStr(strVal);
        sLen = AGetWideStrLen(wideStr);

      WideStr:

        if (len < (A_MIN_SUBSTR_LEN + 1) / 2) {
            if (t->heapPtr + A_COPY_SUBSTR_SIZE > t->heapEnd
                || A_NO_INLINE_ALLOC) {
                *t->tempStack = strVal;
                newWideStr = AAlloc(t, A_COPY_SUBSTR_SIZE);
                if (newWideStr == NULL)
                    return AError;
                strVal = *t->tempStack;
                str = AValueToStr(strVal);
            } else {
                newWideStr = (AWideString *)t->heapPtr;
                t->heapPtr += A_COPY_SUBSTR_SIZE;
            }
            
            AInitNonPointerBlock(&newWideStr->header, sizeof(AWideChar) * len);

            newWideStr->elem[0] = wideStr->elem[begInd];
            newWideStr->elem[1] = wideStr->elem[begInd + 1];
#if A_MIN_SUBSTR_LEN == 9
            if (len > 2) {
                newWideStr->elem[2] = wideStr->elem[begInd + 2];
                newWideStr->elem[3] = wideStr->elem[begInd + 3];
            }
#endif
            return AWideStrToValue(newWideStr);
        } else
            goto SubStr;
    } else
        return ARaiseTypeErrorND(t);

    /* Not reached */

  IndexError:    

    return ARaiseByNum(t, AErrorClassNum[EX_INDEX_ERROR], NULL);
}


/* If v is a character string (length == 1), return the character value of
   the string. Otherwise, return -1 and raise a non-direct exception. */
int ACheckChrStr(AThread *t, AValue v)
{
    if (AIsNarrowStr(v)) {
        if (AGetStrLen(AValueToStr(v)) != 1) {
            ARaiseValueErrorND(t);
            return -1;
        }

        return AValueToStr(v)->elem[0];
    } else if (AIsWideStr(v)) {
        if (AGetWideStrLen(AValueToWideStr(v)) != 1) {
            ARaiseValueErrorND(t);
            return -1;
        }

        return AValueToWideStr(v)->elem[0];
    } else {
        if (!AIsSubStr(v))
            ARaiseTypeErrorND(t);
        else
            ARaiseValueErrorND(t);
        return -1;
    }
    
}


/* Compare two strings lexicographically using the character codes as keys. If
   the return value is < 0, s1 > s2. If the return value is > 0, s1 < s2.
   If the return value is 0, s1 == s2. However, if the return value is
   CMP_FAIL, there was a type error. */
int ACompareStrings(AValue s1, AValue s2)
{
    unsigned char *str1;
    AWideChar *wstr1;
    Assize_t len1;
    
    unsigned char *str2;
    AWideChar *wstr2;
    Assize_t len2;

    if (AIsNarrowStr(s1)) {
        str1 = AValueToStr(s1)->elem;
        len1 = AGetStrLen(AValueToStr(s1));

      Narrow:
        
        if (AIsNarrowStr(s2)) {
            str2 = AValueToStr(s2)->elem;
            len2 = AGetStrLen(AValueToStr(s2));

          NarrowVsNarrow:

            while (len1 > 0 && len2 > 0) {
                if (*str1 != *str2)
                    return *str1 - *str2;
                len1--;
                len2--;
                str1++;
                str2++;
            }

            return len1 - len2;
        }

        if (AIsSubStr(s2)) {
            ASubString *ss;

            ss = AValueToSubStr(s2);
            len2 = AValueToInt(ss->len);

            if (AIsNarrowStr(ss->str)) {
                str2 = AValueToStr(ss->str)->elem + AValueToInt(ss->ind);
                goto NarrowVsNarrow;
            } else {
                wstr2 = AValueToWideStr(ss->str)->elem + AValueToInt(ss->ind);
                goto NarrowVsWide;
            }
        }

        if (AIsWideStr(s2)) {
            wstr2 = AValueToWideStr(s2)->elem;
            len2 = AGetWideStrLen(AValueToWideStr(s2));

          NarrowVsWide:

            while (len1 > 0 && len2 > 0) {
                if (*str1 != *wstr2)
                    return *str1 - *wstr2;
                len1--;
                len2--;
                str1++;
                wstr2++;
            }

            return len1 - len2;
        }

        return A_CMP_FAIL;
    }

    if (AIsSubStr(s1)) {
        ASubString *ss;

        ss = AValueToSubStr(s1);
        len1 = AValueToInt(ss->len);
        
        if (AIsNarrowStr(ss->str)) {
            str1 = AValueToStr(ss->str)->elem + AValueToInt(ss->ind);
            goto Narrow;
        } else {
            wstr1 = AValueToWideStr(ss->str)->elem + AValueToInt(ss->ind);
            goto Wide;
        }
    }

    if (AIsWideStr(s1)) {
        wstr1 = AValueToWideStr(s1)->elem;
        len1 = AGetWideStrLen(AValueToStr(s1));

      Wide:
        
        if (AIsWideStr(s2)) {
            wstr2 = AValueToWideStr(s2)->elem;
            len2 = AGetWideStrLen(AValueToWideStr(s2));

          WideVsWide:
            
            while (len1 > 0 && len2 > 0) {
                if (*wstr1 != *wstr2)
                    return *wstr1 - *wstr2;
                len1--;
                len2--;
                wstr1++;
                wstr2++;
            }

            return len1 - len2;
        }

        if (AIsSubStr(s2)) {
            ASubString *ss;

            ss = AValueToSubStr(s2);
            len2 = AValueToInt(ss->len);

            if (AIsNarrowStr(ss->str)) {
                str2 = AValueToStr(ss->str)->elem + AValueToInt(ss->ind);
                goto WideVsNarrow;
            } else {
                wstr2 = AValueToWideStr(ss->str)->elem + AValueToInt(ss->ind);
                goto WideVsWide;
            }
        }

        if (AIsNarrowStr(s2)) {
            str2 = AValueToStr(s2)->elem;
            len2 = AGetStrLen(AValueToStr(s2));

          WideVsNarrow:

            while (len1 > 0 && len2 > 0) {
                if (*wstr1 != *str2)
                    return *wstr1 - *str2;
                len1--;
                len2--;
                wstr1++;
                str2++;
            }

            return len1 - len2;
        }
    }

    return A_CMP_FAIL;
}


/* Create a string that is formed by concatenating the argument string num
   times. Asssume that strVal is a string. */
AValue ARepeatString(AThread *t, AValue strVal, Assize_t num)
{
    Assize_t len;
    Assize_t ind;

    *t->tempStack = strVal;

    if (AIsNarrowStr(strVal)) {
        AString *s;
        unsigned char *dst;
        unsigned char *src;
        Assize_t size;

        len = AGetStrLen(AValueToStr(strVal));
        ind = 0;

      Narrow:

        size = num * len;
        s = AAlloc(t, sizeof(AValue) + size);
        if (s == NULL)
            return AError;

        AInitNonPointerBlock(&s->header, size);

        if (len > 0) {
            dst = s->elem;
            src = AValueToStr(*t->tempStack)->elem + ind;

            if (len == 1) {
                /* Optimize the case where we are repeating a single
                   character. */
                memset(dst, src[0], num);
            } else {
                while (num > 0) {
                    Assize_t i;
                    
                    i = 0;
                    do {
                        dst[i] = src[i];
                        i++;
                    } while (i < len);
                    
                    dst += len;
                    num--;
                }
            }
        }
        
        return AStrToValue(s);
    }

    if (AIsWideStr(strVal)) {
        AWideString *s;
        AWideChar *dst;
        AWideChar *src;
        unsigned size;

        len = AGetWideStrLen(AValueToWideStr(strVal));
        ind = 0;

      Wide:

        size = num * len * sizeof(AWideChar);
        s = AAlloc(t, sizeof(AValue) + size);
        if (s == NULL)
            return AError;

        AInitNonPointerBlock(&s->header, size);

        if (len > 0) {
            dst = s->elem;
            src = AValueToWideStr(*t->tempStack)->elem + ind;

            while (num > 0) {
                Assize_t i;
                
                i = 0;
                do {
                    dst[i] = src[i];
                    i++;
                } while (i < len);

                dst += len;
                num--;
            }
        }
        
        return AWideStrToValue(s);
    }

    {
        /* Substring */

        ASubString *ss;

        ss = AValueToSubStr(strVal);
        ind = AValueToInt(ss->ind);
        len = AValueToInt(ss->len);
        strVal = ss->str;
        *t->tempStack = strVal;
        
        if (AIsNarrowStr(strVal))
            goto Narrow;
        else
            goto Wide;
    }
}


/* Create a narrow string with the given length. Copy the contents from the
   character buffer unless it is NULL. */
AValue ACreateString(AThread *t, const char *buf, Assize_t len)
{
    AString *str;

    str = AAlloc(t, sizeof(AValue) + len);
    if (str == NULL)
        return AError;

    AInitNonPointerBlock(&str->header, len);
    if (buf != NULL)
        ACopyMem(str->elem, buf, len);

    return AStrToValue(str);
}


/* Create a narrow string from a zero-terminated string. */
AValue ACreateStringFromCStr(AThread *t, const char *str)
{
    return ACreateString(t, str, strlen(str));
}


/* Create a wide/narrow string with the given length. Copy the contents from
   the character buffer. */
AValue ACreateWideString(AThread *t, const AWideChar *buf, Assize_t len)
{
    AWideString *wideStr;
    AString *narrowStr;
    Assize_t i;

    for (i = 0; i < len; i++) {
        if (buf[i] > 0xff)
            break;
    }

    if (i < len) {
        wideStr = AAlloc(t, sizeof(AValue) + sizeof(AWideChar) * len);
        if (wideStr == NULL)
            return AError;

        AInitNonPointerBlock(&wideStr->header, sizeof(AWideChar) * len);
        ACopyMem(wideStr->elem, buf, sizeof(AWideChar) * len);

        return AWideStrToValue(wideStr);
    } else {
        narrowStr = AAlloc(t, sizeof(AValue) + len);
        if (narrowStr == NULL)
            return AError;

        AInitNonPointerBlock(&narrowStr->header, len);
        for (i = 0; i < len; i++)
            narrowStr->elem[i] = buf[i];

        return AStrToValue(narrowStr);
    }
}


enum { BUF_STATIC, BUF_DYNAMIC };


/* Copy the contents of a string to a buffer as a C string.
   Return -1 if failed and raise an exception. Return the length of the string
   if successful. */
Assize_t AGetCStrFromString(AThread *t, AValue strVal, char *buf,
                            Assize_t bufLen)
{
    /* FIX: do not allocate memory?? */
    
    AString *str;
    Assize_t len;
    
    strVal = ANormalizeNarrowString(t, strVal);
    if (AIsError(strVal))
        return -1;

    str = AValueToStr(strVal);
    len = AGetStrLen(str);

    if (len + 1 <= bufLen) {
        ACopyMem(buf, str->elem, len);
        buf[len] = '\0';
        return len;
    }
    
    ARaiseValueErrorND(t);
    return -1;
}


/* Allocate a buffer and copy the contents of a string to it as a C string.
   If buf != NULL, use buf to hold the contents if possible. Return a pointer
   to the buffer or NULL if unsuccessful. In that case, raise an exception. */
char *AAllocCStrFromString(AThread *t, AValue strVal, char *buf,
                           Assize_t bufLen)
{
    AString *str;
    Assize_t len;
    
    strVal = ANormalizeNarrowString(t, strVal);
    if (AIsError(strVal))
        return NULL;

    str = AValueToStr(strVal);
    len = AGetStrLen(str);

    if (len + 2 <= bufLen)
        buf[0] = BUF_STATIC;
    else {
        t->tempStack[0] = strVal;
        buf = AAllocStatic(AGetBlockSize(len + 2));
        if (buf == NULL) {
            ARaiseMemoryErrorND(t);
            return NULL;
        }
        buf[0] = BUF_DYNAMIC;
        str = AValueToStr(t->tempStack[0]);
    }

    buf++;
    ACopyMem(buf, str->elem, len);
    buf[len] = '\0';

    /* Did the original string have zero characters? */
    if (strlen(buf) < len) {
        ARaiseValueErrorND(t);
        return NULL;
    }

    return buf;
}


/* Free the buffer returned by AllocCStrFromString unless the caller-supplied
   buffer was used. */
void AFreeCStrFromString(const char *str)
{
    if (str[-1] == BUF_DYNAMIC)
        AFreeStatic((void *)(str - 1));
}


Assize_t AStrLen(AValue v)
{
    if (AIsNarrowStr(v))
        return AGetStrLen(AValueToStr(v));
    else if (AIsWideStr(v))
        return AGetWideStrLen(AValueToWideStr(v));
    else
        return AGetSubStrLen(AValueToSubStr(v));
}


AWideChar AStrItem_(AValue v, Assize_t index)
{
    if (AIsWideStr(v))
        return AGetWideStrElem(v)[index];
    else {
        ASubString *subStr = AValueToSubStr(v);
        if (AIsWideStr(subStr->str))
            return AGetWideSubStrElem(v)[index];
        else
            return AGetSubStrElem(v)[index];
    }
}


void ASetStrItem(AValue v, Assize_t index, AWideChar ch)
{
    if (AIsNarrowStr(v))
        AGetStrElem(v)[index] = ch;
    else if (AIsWideStr(v))
        AGetWideStrElem(v)[index] = ch;
    else {
        ASubString *subStr = AValueToSubStr(v);
        if (AIsWideStr(subStr->str))
            AGetWideSubStrElem(v)[index] = ch;
        else
            AGetSubStrElem(v)[index] = ch;
    }
}


AValue AConcatStringAndCStr(AThread *t, AValue str, const char *cstr)
{
    AValue cstrVal;
    
    *t->tempStack = str;
    cstrVal = ACreateStringFromCStr(t, cstr);
    if (AIsError(cstrVal))
        return cstrVal;

    return AConcatStrings(t, *t->tempStack, cstrVal); 
}


/* Check if substr is included in a string. Assume str is string, but substr
   can be of any value. Return Int value 0 if the string could not be found,
   Int value 1 if it could be found. Raise a non-direct exception and return
   AError if there was a type error. */
AValue AIsInString(AThread *t, AValue str, AValue substr)
{
    if (!AIsStr(substr))
        return ARaiseTypeErrorND(t);

    return Find(str, 0, substr) >= 0 ? AIntToValue(1) : AZero;
}


AValue AStringHashValue(AValue str)
{
    Assize_t len;
    unsigned char *s;
    AWideChar *ws;
    Assize_t i;
    unsigned hashValue = 0;

    len = AStrLen(str);
    
    if (AIsNarrowStr(str)) {
        s = AValueToStr(str)->elem;
        for (i = 0; i < len; i++)
            hashValue += hashValue * 32 + s[i];
    } else if (AIsWideStr(str)) {
        ws = AValueToWideStr(str)->elem;
        for (i = 0; i < len; i++)
            hashValue += hashValue * 32 + ws[i];
    } else {
        ASubString *subStr = AValueToSubStr(str);
        if (AIsWideStr(subStr->str)) {
            ws = AGetWideSubStrElem(str);
            for (i = 0; i < len; i++)
                hashValue += hashValue * 32 + ws[i];
        } else {
            s = AGetSubStrElem(str);
            for (i = 0; i < len; i++)
                hashValue += hashValue * 32 + s[i];
        }
    }

    return AIntToValue(hashValue);
}


AValue AStrRepr(AThread *t, AValue *strPtr)
{
    AValue str;
    Assize_t len;
    AValue resultStr;
    Assize_t resultLen;
    Assize_t i;
    Assize_t ri;
    unsigned char *resultElem;

    str = *strPtr;
    len = AStrLen(str);
    resultLen = 2;

    for (i = 0; i < len; i++) {
        if (AStrItem(str, i) < 32 || AStrItem(str, i) >= 127)
            resultLen += 6;
        else if (AStrItem(str, i) == '"')
            resultLen += 2;
        else
            resultLen++;
    }

    resultStr = AMakeEmptyStr(t, resultLen);

    resultElem = AGetStrElem(resultStr);
    resultElem[0] = '"';
    ri = 1;

    for (i = 0; i < len; i++) {
        AWideChar ch = AStrItem(str, i);
        if (ch < 32 || ch >= 127) {
            char buf[7];
            sprintf(buf, "\\u%04x", ch);
            ACopyMem(resultElem + ri, buf, 6);
            ri += 6;
        } else {
            resultElem[ri++] = ch;
            if (ch == '"')
                resultElem[ri++] = ch;
        }
    }

    resultElem[ri] = '"';
    return resultStr;
}


/* Find leftmost instance of sub-string in a string starting from the specified
   index (0 == start of string). Return -1 if not found, otherwise return the
   index at which the substring was found. Assume str and substr are Str
   values.*/
static Assize_t Find(AValue str, Assize_t i, AValue substr)
{
    Assize_t len1;
    Assize_t len2;
    int ch;
    Assize_t j;
    
    len1 = AStrLen(str);
    len2 = AStrLen(substr);

    if (len2 == 0) {
        if (i <= len1)
            return i;
        else
            return -1;
    }

    /* Get the string start character. */
    ch = AStrItem(substr, 0);

    for (; i <= len1 - len2; i++) {
        if (AStrItem(str, i) == ch) {
            /* Potential match. Verify it. */
            for (j = 1; j < len2; j++) {
                if (AStrItem(str, i + j) != AStrItem(substr, j))
                   break;
            }
            /* Found a match? */
            if (j >= len2)
                return i;
        }
    }
    
    return -1;
}


static AValue SymbolStr(AThread *t, ASymbolInfo *sym)
{
    char buf[1025];
    if (sym == NULL)
        strcpy(buf, "nil");
    else
        AFormatMessage(buf, 1025, "%q", sym);
    /* FIX: Don't use a fixed-size buffer. */
    return AMakeStr(t, buf);
}


/* If buf points to a non-standard string representation of a floating point
   (-)infinity or NaN, convert it to the standard representation (inf, -inf or
   nan). Note that here "standard" refers to standard Alore behaviour, not
   standard C behaviour... */
void AFixInfAndNanStrings(char *buf)
{
    if (buf[0] == 'I' && strcmp(buf, "Infinity") == 0)
        strcpy(buf, "inf");
    else if (buf[0] == 'N' && strcmp(buf, "NaN") == 0)
        strcpy(buf, "nan");
    else if (buf[0] == '-') {
        if (buf[1] == 'I' && strcmp(buf, "-Infinity") == 0)
            strcpy(buf, "-inf");
        else if (buf[1] == 'N' && strcmp(buf, "-NaN") == 0)
            strcpy(buf, "nan");
        else if (buf[1] == 'n' && strcmp(buf, "-nan") == 0)
            strcpy(buf, "nan");
    }
#ifdef A_HAVE_WINDOWS
    /* At least mingw has rather peculiar representations for inifinities and
       nans. */
    if (buf[0] == '1' && buf[1] == '.' && buf[2] == '#') {
        if (strcmp(buf, "1.#INF") == 0)
            strcpy(buf, "inf");
        else if (strcmp(buf, "1.#IND") == 0)
            strcpy(buf, "nan");
    } else if (buf[0] == '-' && buf[1] == '1' && buf[2] == '.'
               && buf[3] == '#') {
        if (strcmp(buf, "-1.#INF") == 0)
            strcpy(buf, "-inf");
        else if (strcmp(buf, "-1.#IND") == 0)
            strcpy(buf, "nan");
    }
#endif
}


AValue AStrReplace(AThread *t, AValue *frame)
{
    Assize_t origLen, newLen;
    Assize_t srcLen, dstLen;
    Assize_t i, j;
    Assize_t max;
    int ch;
    Assize_t n;

    frame[0] = AWrapObject(t, frame[0], -1);
    
    AExpectStr(t, frame[1]);
    AExpectStr(t, frame[2]);

    if (AIsDefault(frame[3]))
        max = A_SSIZE_T_MAX;
    else
        max = AGetInt(t, frame[3]);
    
    frame[3] = A_UNWRAP(frame[0]);
    origLen = AStrLen(frame[3]);
    srcLen = AStrLen(frame[1]);
    dstLen = AStrLen(frame[2]);

    /* Do nothing if the source string is empty. */
    if (srcLen == 0 || max <= 0)
        return frame[3];

    /* Calculate the length of the new string. */
    newLen = origLen;
    if (srcLen != dstLen) {
        Assize_t c = Count(frame[3], frame[1], max);
        Assize_t x;
        if (c == 0)
            return frame[3];

        /* Perform the calculation. Verify that there is no overflow. */
        x = (dstLen - srcLen) * c;
        if (x / c != dstLen - srcLen)
            return ARaiseRuntimeError(t, "Replace result too long");
        newLen += x;
        if (newLen < 0)
            return ARaiseRuntimeError(t, "Replace result too long");
    }
    
    /* Setup the new string. */
    if ((AIsNarrowStr(frame[3]) || AIsNarrowSubStr(frame[3]))
        && (AIsNarrowStr(frame[1]) || AIsNarrowSubStr(frame[1]))
        && (AIsNarrowStr(frame[2]) || AIsNarrowSubStr(frame[2])))
        frame[4] = AMakeEmptyStr(t, newLen);
    else
        frame[4] = AMakeEmptyStrW(t, newLen);

    i = j = 0;
    ch = AStrItem(frame[1], 0);
    n = 0;
    while (i <= origLen - srcLen && n < max) {
        Assize_t i0 = i;

        while (i <= origLen - srcLen
               && AStrItem(frame[3], i) != ch)
            i++;

        ACopySubStr(frame[4], j, frame[3], i0, i - i0);
        j += i - i0;

        if (i > origLen - srcLen)
            break;
        
        if (IsMatchAt(frame[3], i, frame[1], srcLen)) {
            Assize_t k;
            for (k = 0; k < dstLen; k++) {
                ASetStrItem(frame[4], j, AStrItem(frame[2], k));
                j++;
            }
            i += srcLen;
            n++;
        } else {
            ASetStrItem(frame[4], j, ch);
            i++;
            j++;
        }
    }

    ACopySubStr(frame[4], j, frame[3], i, origLen - i);
    
    return frame[4];
}


static ABool IsMatchAt(AValue src, Assize_t i, AValue dst, Assize_t dstLen)
{
    Assize_t j;
    for (j = 1; j < dstLen; j++) {
        if (AStrItem(src, i + j) !=
            AStrItem(dst, j))
            return 0;
    }
    return 1;
}


static Assize_t Count(AValue str, AValue substr, Assize_t max)
{
    Assize_t len1 = AStrLen(str);
    Assize_t len2 = AStrLen(substr);
    Assize_t count = 0;
    Assize_t i = 0;
    Assize_t j;
    int ch;

    ch = AStrItem(substr, 0);

    while (i <= len1 - len2) {
        if (AStrItem(str, i) == ch) {
            /* Potential match. Verify it. */
            for (j = 1; j < len2; j++) {
                if (AStrItem(str, i + j) !=
                    AStrItem(substr, j))
                   break;
            }
            /* Found a match? */
            if (j >= len2) {
                count++;
                if (count == max)
                    return count;
                i += len2;
            } else
                i++;
        } else
            i++;
    }
    return count;
}


#define IS_SPLIT_SPACE(ch) \
    ((ch) == ' ' || (ch) == '\n' || (ch) == '\r' || (ch) == '\t')


AValue AStrSplit(AThread *t, AValue *frame)
{
    ssize_t strLen;
    ssize_t i, i0;
    ssize_t max;
    ssize_t n;
    AValue ss;
    
    frame[3] = frame[0];
    frame[0] = AWrapObject(t, frame[0], -1);

    frame[4] = AMakeArray(t, 0);

    strLen = AStrLen(frame[3]);
    if (!AIsDefault(frame[2]))
        max = AGetInt(t, frame[2]);
    else
        max = A_SSIZE_T_MAX;
    
    n = 0;
    
    if (frame[1] != ADefault && frame[1] != ANil) {
        /* Explicit separator string */
        ssize_t sepLen;
        int ch;
        
        AExpectStr(t, frame[1]);
        
        sepLen = AStrLen(frame[1]);
        if (sepLen == 0)
            return ARaiseValueError(t, "Empty separator");
        
        ch = AStrItem(frame[1], 0);
        i0 = 0;
        for (i = 0; i <= strLen - sepLen;) {
            if (AStrItem(frame[3], i) == ch
                && IsMatchAt(frame[3], i, frame[1], sepLen) && n < max) {
                ss = ASubStr(t, frame[3], i0, i);
                AAppendArray(t, frame[4], ss);
                i += sepLen;
                i0 = i;
                n++;
            } else
                i++;
        }
    } else {
        /* Run of whitespace as a separator. Whitespace at the start and the
           end of the string is ignored. */
        for (i0 = 0; i0 < strLen; i0++) {
            int ch = AStrItem(frame[3], i0);
            if (!IS_SPLIT_SPACE(ch))
                break;
        }
        if (i0 >= strLen)
            goto End;
        for (i = i0; i < strLen;) {
            int ch = AStrItem(frame[3], i);
            if (IS_SPLIT_SPACE(ch) && n < max) {
                ss = ASubStr(t, frame[3], i0, i);
                AAppendArray(t, frame[4], ss);
                for (i++; i < strLen; i++) {
                    ch = AStrItem(frame[3], i);
                    if (!IS_SPLIT_SPACE(ch))
                        break;
                }
                if (i >= strLen)
                    goto End;
                n++;
                i0 = i;
            } else
                i++;
        }
    }

    ss = ASubStr(t, frame[3], i0, strLen);
    AAppendArray(t, frame[4], ss);

  End:
    
    return frame[4];
}


AValue AStrJoin(AThread *t, AValue *frame)
{
    ssize_t len;
    ssize_t strLen;
    ssize_t i;
    ssize_t ti;
    ABool isWide;
    ssize_t sepLen;

    frame[3] = frame[0];
    frame[0] = AWrapObject(t, frame[0], -1);

    sepLen = AStrLen(frame[3]);
    isWide = AIsWideStr(frame[3]) || AIsWideSubStr(frame[3]);

    /* Calculate the total length of the result. */
    len = ALen(t, frame[1]);
    if (len > 0)
        strLen = (len - 1) * sepLen;
    else if (len < 0)
        return AError;
    else
        strLen = 0;
    for (i = 0; i < len; i++) {
        AValue s = AGetItemAt(t, frame[1], i);
        if (!AIsNarrowStr(s) && !AIsNarrowSubStr(s)) {
            if (AIsStr(s))
                isWide = TRUE;
            else if (AIsError(s))
                return AError;
            else
                ARaiseTypeError(t, AMsgStrExpected);
        }
        strLen += AStrLen(s);
    }

    /* Optimize for the special case of a single-element sequence. */
    if (len == 1)
        return AGetItemAt(t, frame[1], 0);
    
    /* Create a new string. */
    if (isWide)
        frame[2] = AMakeEmptyStrW(t, strLen);
    else
        frame[2] = AMakeEmptyStr(t, strLen);
    
    /* Copy strings. */
    ti = 0;
    for (i = 0; i < len; i++) {
        ssize_t slen;
        AValue s = AGetItemAt(t, frame[1], i);
        
        if (!AIsStr(s)) {
            if (AIsError(s))
                return AError;
            else
                ARaiseTypeError(t, AMsgStrExpected);
        }

        slen = AStrLen(s);
        ACopySubStr(frame[2], ti, s, 0, slen);
        ti += slen;

        if (sepLen > 0 && i < len - 1) {
            ACopySubStr(frame[2], ti, frame[3], 0, sepLen);
            ti += sepLen;
        }
    }
    
    return frame[2];
}


AValue AJoin(AThread *t, AValue sequence, AValue separator)
{
    AValue *frame = AAllocTemps(t, 4);
    AValue result;

    frame[1] = sequence;
    if (separator != ADefault)
        frame[0] = separator;
    else
        frame[0] = AMakeStr(t, "");
    result = AStrJoin(t, frame);

    AFreeTemps(t, 4);

    return result;
}


void ACopySubStr(AValue dst, ssize_t dstIndex, AValue src, ssize_t srcIndex,
                 ssize_t srcLen)
{
    if (AIsNarrowStr(dst)) {
        /* narrow -> narrow (assume src is narrow) */
        const unsigned char *srcItems = ANarrowStrItems(src);
        unsigned char *dstItems = AGetStrElem(dst);
        memcpy(dstItems + dstIndex, srcItems + srcIndex, srcLen);
    } else {
        AWideChar *dstItems = AGetWideStrElem(dst);
        
        if (AIsWideStr(src) || AIsWideSubStr(src)) {
            /* wide -> wide */
            const AWideChar *srcItems = AWideStrItems(src);
            memcpy(dstItems + dstIndex, srcItems + srcIndex,
                   srcLen * sizeof(AWideChar));
        } else {
            /* narrow -> wide */
            ssize_t i;
            const unsigned char *srcItems = ANarrowStrItems(src);
            for (i = 0; i < srcLen; i++)
                dstItems[dstIndex + i] = srcItems[srcIndex + i];
        }
    }
}


AValue AStrIter(AThread *t, AValue *frame)
{
    frame[1] = frame[0];
    frame[0] = AWrapObject(t, frame[0], -1);
    
    return ACallValue(t, AGlobalByNum(AStrIterNum), 1, frame + 1);
}


AValue AStrIterCreate(AThread *t, AValue *frame)
{
    AValue len;
    
    AExpectStr(t, frame[1]);
    ASetMemberDirect(t, frame[0], STR_ITER_I, AZero);
    len = AIntToValue(AStrLen(frame[1]));
    ASetMemberDirect(t, frame[0], STR_ITER_LEN, len);
    ASetMemberDirect(t, frame[0], STR_ITER_S, frame[1]);
    return frame[0];
}


AValue AStrIterHasNext(AThread *t, AValue *frame)
{
    AValue i = AMemberDirect(frame[0], STR_ITER_I);
    AValue n = AMemberDirect(frame[0], STR_ITER_LEN);
    if (i < n)
        return ATrue;
    else
        return AFalse;
}


AValue AStrIterNext(AThread *t, AValue *frame)
{
    AValue i = AMemberDirect(frame[0], STR_ITER_I);
    AValue n = AMemberDirect(frame[0], STR_ITER_LEN);

    if (i < n) {
        int ch;
        
        ch = AStrItem(AMemberDirect(frame[0], STR_ITER_S), AValueToInt(i));
        
        i += AIntToValue(1);
        AValueToInstance(frame[0])->member[STR_ITER_I] = i;
        
        return AMakeCh(t, ch);
    } else
        return ARaiseValueError(t, "No items left");
}
